#!/usr/bin/env python3
#
# See the impact of a Futhark commit compared to the previous one we
# have benchmarking for.

import sys
import subprocess
from urllib.request import urlopen
from urllib.error import HTTPError
import json
import tempfile
import os
import gzip


def url_for(backend, system, commit):
    return (
        "https://futhark-lang.org/benchmark-results/futhark-"
        "{}-{}-{}.json.gz"
    ).format(backend, system, commit)


def results_for_commit(backend, system, commit):
    try:
        url = url_for(backend, system, commit)
        print("Fetching {}...".format(url))
        return json.loads(gzip.decompress(urlopen(url).read()))
    except HTTPError:
        return None


def first_commit_with_results(backend, system, commits):
    for commit in commits:
        res = results_for_commit(backend, system, commit)
        if res:
            return commit, res


def find_commits(start):
    return (
        subprocess.check_output(["git", "rev-list", start])
        .decode("utf-8")
        .splitlines()
    )


if __name__ == "__main__":
    backend, system, commit = sys.argv[1:4]

    now = results_for_commit(backend, system, commit)

    if not now:
        print("No results found")
        sys.exit(1)

    if len(sys.argv) == 5:
        commits = find_commits(sys.argv[4])
    else:
        commits = find_commits(commit)[1:]

    then_commit, then = first_commit_with_results(backend, system, commits[1:])

    print("Comparing {}".format(commit))
    print("     with {}".format(then_commit))

    with tempfile.NamedTemporaryFile(prefix=commit, mode="w") as now_file:
        with tempfile.NamedTemporaryFile(
            prefix=then_commit, mode="w"
        ) as then_file:
            json.dump(now, now_file)
            json.dump(then, then_file)
            now_file.flush()
            then_file.flush()
            os.system(
                "futhark benchcmp {} {}".format(then_file.name, now_file.name)
            )



# Start of memory.py.

import ctypes as ct


def allocateMem(size):
    return np.empty(size, dtype=np.byte)


# Copy an array if its is not-None.  This is important for treating
# Numpy arrays as flat memory, but has some overhead.
def normaliseArray(x):
    if (x.base is x) or (x.base is None):
        return x
    else:
        return x.copy()


def unwrapArray(x):
    return x.ravel().view(np.byte)


def indexArray(x, offset, bt):
    return x.view(bt)[offset]


def writeScalarArray(x, offset, v):
    x.view(type(v))[offset] = v


# An opaque Futhark value.
class opaque(object):
    def __init__(self, desc, *payload):
        self.data = payload
        self.desc = desc

    def __repr__(self):
        return "<opaque Futhark value of type {}>".format(self.desc)


# LMAD stuff


def lmad_contiguous_search(checked, expected, strides, shape, used):
    for i in range(len(strides)):
        for j in range(len(strides)):
            if not used[j] and strides[j] == expected and strides[j] >= 0:
                used[j] = True
                if checked + 1 == len(strides) or lmad_contiguous_search(
                    checked + 1, expected * shape[j], strides, shape, used
                ):
                    return True
                used[j] = False
    return False


def lmad_contiguous(strides, shape):
    used = len(strides) * [False]
    return lmad_contiguous_search(0, 1, strides, shape, used)


def lmad_memcpyable(dst_strides, src_strides, shape):
    if not lmad_contiguous(dst_strides, shape):
        return False
    for i in range(len(dst_strides)):
        if dst_strides[i] != src_strides[i] and shape[i] != 1:
            return False
    return True


def lmad_is_tr(strides, shape):
    r = len(shape)
    for i in range(1, r):
        n = 1
        m = 1
        ok = True
        expected = 1
        # Check strides before 'i'.
        for j in range(i - 1, -1, -1):
            ok = ok and strides[j] == expected
            expected *= shape[j]
            n *= shape[j]
        # Check strides after 'i'.
        for j in range(r - 1, i - 1, -1):
            ok = ok and strides[j] == expected
            expected *= shape[j]
            m *= shape[j]
        if ok:
            return (n, m)
    return None


def lmad_map_tr(dst_strides, src_strides, shape):
    r = len(dst_strides)
    rowmajor_strides = [0] * r
    rowmajor_strides[r - 1] = 1

    for i in range(r - 2, -1, -1):
        rowmajor_strides[i] = rowmajor_strides[i + 1] * shape[i + 1]

    # map_r will be the number of mapped dimensions on top.
    map_r = 0
    k = 1
    for i in range(r):
        if (
            dst_strides[i] != rowmajor_strides[i]
            or src_strides[i] != rowmajor_strides[i]
        ):
            break
        else:
            k *= shape[i]
            map_r += 1

    if rowmajor_strides[map_r:] == dst_strides[map_r:]:
        r = lmad_is_tr(src_strides[map_r:], shape[map_r:])
        if r is not None:
            (n, m) = r
            return (k, n, m)
    elif rowmajor_strides[map_r:] == src_strides[map_r:]:
        r = lmad_is_tr(dst_strides[map_r:], shape[map_r:])
        if r is not None:
            (n, m) = r
            return (k, m, n)  # Sic!
    return None


def lmad_copy_elements(
    pt, dst, dst_offset, dst_strides, src, src_offset, src_strides, shape
):
    if len(shape) == 1:
        for i in range(shape[0]):
            writeScalarArray(
                dst,
                dst_offset + i * dst_strides[0],
                indexArray(src, src_offset + i * src_strides[0], pt),
            )
    else:
        for i in range(shape[0]):
            lmad_copy_elements(
                pt,
                dst,
                dst_offset + i * dst_strides[0],
                dst_strides[1:],
                src,
                src_offset + i * src_strides[0],
                src_strides[1:],
                shape[1:],
            )


def lmad_copy(
    pt, dst, dst_offset, dst_strides, src, src_offset, src_strides, shape
):
    if lmad_memcpyable(dst_strides, src_strides, shape):
        dst[
            dst_offset * ct.sizeof(pt) : dst_offset * ct.sizeof(pt)
            + np.prod(shape) * ct.sizeof(pt)
        ] = src[
            src_offset * ct.sizeof(pt) : src_offset * ct.sizeof(pt)
            + np.prod(shape) * ct.sizeof(pt)
        ]
    else:
        lmad_copy_elements(
            pt,
            dst,
            dst_offset,
            dst_strides,
            src,
            src_offset,
            src_strides,
            shape,
        )


# End of memory.py.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Futhark documentation build configuration file, created by
# sphinx-quickstart on Tue Mar 24 14:21:12 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import re
from pygments.lexer import RegexLexer, bygroups
from pygments import token
from pygments import unistring as uni
from pygments.token import (
    Text,
    Comment,
    Operator,
    Keyword,
    Name,
    String,
    Number,
    Punctuation,
    Whitespace,
)
from sphinx.highlighting import lexers
from typing import Dict, Any, List, Tuple

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.todo", "sphinx.ext.mathjax"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Futhark"
copyright = "2013-2020, DIKU, University of Copenhagen"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.


# No reason for a cabal file parser; let's just hack it.
def get_version():
    # Get cabal file
    cabal_file = open("../futhark.cabal", "r").read()
    # Extract version
    return re.search(
        r"^version:[ ]*([^ ]*)$", cabal_file, flags=re.MULTILINE
    ).group(1)


version = get_version()
# The full version, including alpha/beta/rc tags.
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "lib"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = false

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


class FutharkLexer(RegexLexer):
    """
    A Futhark lexer

    .. versionadded:: 2.8
    """

    name = "Futhark"
    url = "https://futhark-lang.org/"
    aliases = ["futhark"]
    filenames = ["*.fut"]
    mimetypes = ["text/x-futhark"]

    num_types = (
        "i8",
        "i16",
        "i32",
        "i64",
        "u8",
        "u16",
        "u32",
        "u64",
        "f32",
        "f64",
    )

    other_types = ("bool",)

    reserved = (
        "if",
        "then",
        "else",
        "def",
        "let",
        "loop",
        "in",
        "with",
        "type",
        "val",
        "entry",
        "for",
        "while",
        "do",
        "case",
        "match",
        "include",
        "import",
        "module",
        "open",
        "local",
        "assert",
        "_",
    )

    ascii = (
        "NUL",
        "SOH",
        "[SE]TX",
        "EOT",
        "ENQ",
        "ACK",
        "BEL",
        "BS",
        "HT",
        "LF",
        "VT",
        "FF",
        "CR",
        "S[OI]",
        "DLE",
        "DC[1-4]",
        "NAK",
        "SYN",
        "ETB",
        "CAN",
        "EM",
        "SUB",
        "ESC",
        "[FGRU]S",
        "SP",
        "DEL",
    )

    num_postfix = r"(%s)?" % "|".join(num_types)

    identifier_re = "[a-zA-Z_][a-zA-Z_0-9']*"

    # opstart_re = '+\-\*/%=\!><\|&\^'

    tokens = {
        "root": [
            (r"--(.*?)$", Comment.Single),
            (r"\s+", Whitespace),
            (r"\(\)", Punctuation),
            (r"\b(%s)(?!\')\b" % "|".join(reserved), Keyword.Reserved),
            (
                r"\b(%s)(?!\')\b" % "|".join(num_types + other_types),
                Keyword.Type,
            ),
            # Identifiers
            (r"#\[([a-zA-Z_\(\) ]*)\]", Comment.Preproc),
            (r"[#!]?(%s\.)*%s" % (identifier_re, identifier_re), Name),
            (r"\\", Operator),
            (r"[-+/%=!><|&*^][-+/%=!><|&*^.]*", Operator),
            (r"[][(),:;`{}?.\']", Punctuation),
            #  Numbers
            (
                r"0[xX]_*[\da-fA-F](_*[\da-fA-F])*_*[pP][+-]?\d(_*\d)*"
                + num_postfix,
                Number.Float,
            ),
            (
                r"0[xX]_*[\da-fA-F](_*[\da-fA-F])*\.[\da-fA-F](_*[\da-fA-F])*"
                r"(_*[pP][+-]?\d(_*\d)*)?" + num_postfix,
                Number.Float,
            ),
            (r"\d(_*\d)*_*[eE][+-]?\d(_*\d)*" + num_postfix, Number.Float),
            (
                r"\d(_*\d)*\.\d(_*\d)*(_*[eE][+-]?\d(_*\d)*)?" + num_postfix,
                Number.Float,
            ),
            (r"0[bB]_*[01](_*[01])*" + num_postfix, Number.Bin),
            (r"0[xX]_*[\da-fA-F](_*[\da-fA-F])*" + num_postfix, Number.Hex),
            (r"\d(_*\d)*" + num_postfix, Number.Integer),
            #  Character/String Literals
            (r"'", String.Char, "character"),
            (r'"', String, "string"),
            #  Special
            (r"\[[a-zA-Z_\d]*\]", Keyword.Type),
            (r"\(\)", Name.Builtin),
        ],
        "character": [
            # Allows multi-chars, incorrectly.
            (r"[^\\']'", String.Char, "#pop"),
            (r"\\", String.Escape, "escape"),
            ("'", String.Char, "#pop"),
        ],
        "string": [
            (r'[^\\"]+', String),
            (r"\\", String.Escape, "escape"),
            ('"', String, "#pop"),
        ],
        "escape": [
            (r'[abfnrtv"\'&\\]', String.Escape, "#pop"),
            (r"\^[][" + uni.Lu + r"@^_]", String.Escape, "#pop"),
            ("|".join(ascii), String.Escape, "#pop"),
            (r"o[0-7]+", String.Escape, "#pop"),
            (r"x[\da-fA-F]+", String.Escape, "#pop"),
            (r"\d+", String.Escape, "#pop"),
            (r"(\s+)(\\)", bygroups(Whitespace, String.Escape), "#pop"),
        ],
    }


lexers["futhark"] = FutharkLexer()

highlight_language = "text"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = false


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "futhark"
html_theme_path = ["_theme"]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options: Dict[Any, Any] = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    "**": [
        "globaltoc.html",
        "relations.html",
        "sourcelink.html",
        "searchbox.html",
    ]
}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = false

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "Futharkdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements: Dict[str, str] = {
    # The paper size ('letterpaper' or 'a4paper').
    #'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    ("index", "Futhark.tex", "Futhark User's Guide", "DIKU", "manual"),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = false

# If true, show page references after internal links.
# latex_show_pagerefs = false

# If true, show URL addresses after external links.
# latex_show_urls = false

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages: List[Tuple[str, str, str, List[Any], int]] = [
    ("man/futhark", "futhark", "a parallel functional array language", [], 1),
    (
        "man/futhark-autotune",
        "futhark-autotune",
        "calibrate run-time parameters",
        [],
        1,
    ),
    ("man/futhark-c", "futhark-c", "compile Futhark to sequential C", [], 1),
    (
        "man/futhark-multicore",
        "futhark-multicore",
        "compile Futhark to multithreaded C",
        [],
        1,
    ),
    (
        "man/futhark-ispc",
        "futhark-ispc",
        "compile Futhark to multithreaded ISPC",
        [],
        1,
    ),
    (
        "man/futhark-opencl",
        "futhark-opencl",
        "compile Futhark to OpenCL",
        [],
        1,
    ),
    ("man/futhark-cuda", "futhark-cuda", "compile Futhark to CUDA", [], 1),
    ("man/futhark-hip", "futhark-hip", "compile Futhark to HIP", [], 1),
    (
        "man/futhark-python",
        "futhark-python",
        "compile Futhark to sequential Python",
        [],
        1,
    ),
    (
        "man/futhark-pyopencl",
        "futhark-pyopencl",
        "compile Futhark to Python and OpenCL",
        [],
        1,
    ),
    (
        "man/futhark-wasm",
        "futhark-wasm",
        "compile Futhark to WebAssembly",
        [],
        1,
    ),
    (
        "man/futhark-wasm-multicore",
        "futhark-wasm-multicore",
        "compile Futhark to parallel WebAssembly",
        [],
        1,
    ),
    ("man/futhark-run", "futhark-run", "interpret Futhark program", [], 1),
    (
        "man/futhark-repl",
        "futhark-repl",
        "interactive Futhark read-eval-print-loop",
        [],
        1,
    ),
    ("man/futhark-test", "futhark-test", "test Futhark programs", [], 1),
    (
        "man/futhark-bench",
        "futhark-bench",
        "benchmark Futhark programs",
        [],
        1,
    ),
    (
        "man/futhark-doc",
        "futhark-doc",
        "generate documentation for Futhark code",
        [],
        1,
    ),
    (
        "man/futhark-dataset",
        "futhark-dataset",
        "generate random data sets",
        [],
        1,
    ),
    (
        "man/futhark-fmt",
        "futhark-fmt",
        "format Futhark programs",
        [],
        1,
    ),
    ("man/futhark-pkg", "futhark-pkg", "manage Futhark packages", [], 1),
    (
        "man/futhark-literate",
        "futhark-literate",
        "execute literate Futhark program",
        [],
        1,
    ),
    (
        "man/futhark-script",
        "futhark-script",
        "execute FutharkScript expression",
        [],
        1,
    ),
    (
        "man/futhark-profile",
        "futhark-profile",
        "profile Futhark programs",
        [],
        1,
    ),
]

# If true, show URL addresses after external links.
# man_show_urls = false


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "Futhark",
        "Futhark Documentation",
        "DIKU",
        "Futhark",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = false



# Stub code for OpenCL setup.

import pyopencl as cl
import numpy as np
import sys

if cl.version.VERSION < (2015, 2):
    raise Exception(
        "Futhark requires at least PyOpenCL version 2015.2.  Installed version is %s."
        % cl.version.VERSION_TEXT
    )

TR_BLOCK_DIM = 16
TR_TILE_DIM = TR_BLOCK_DIM * 2
TR_ELEMS_PER_THREAD = 8


def parse_preferred_device(s):
    pref_num = 0
    if len(s) > 1 and s[0] == "#":
        i = 1
        while i < len(s):
            if not s[i].isdigit():
                break
            else:
                pref_num = pref_num * 10 + int(s[i])
            i += 1
        while i < len(s) and s[i].isspace():
            i += 1
        return (s[i:], pref_num)
    else:
        return (s, 0)


def get_prefered_context(
    interactive=False, platform_pref=None, device_pref=None
):
    if device_pref != None:
        (device_pref, device_num) = parse_preferred_device(device_pref)
    else:
        device_num = 0

    if interactive:
        return cl.create_some_context(interactive=True)

    def blacklisted(p, d):
        return (
            platform_pref == None
            and device_pref == None
            and p.name == "Apple"
            and d.name.find("Intel(R) Core(TM)") >= 0
        )

    def platform_ok(p):
        return not platform_pref or p.name.find(platform_pref) >= 0

    def device_ok(d):
        return not device_pref or d.name.find(device_pref) >= 0

    device_matches = 0

    for p in cl.get_platforms():
        if not platform_ok(p):
            continue
        for d in p.get_devices():
            if blacklisted(p, d) or not device_ok(d):
                continue
            if device_matches == device_num:
                return cl.Context(devices=[d])
            else:
                device_matches += 1
    raise Exception(
        "No OpenCL platform and device matching constraints found."
    )


def param_assignment(s):
    name, value = s.split("=")
    return (name, int(value))


def check_types(self, required_types):
    if "f64" in required_types:
        if (
            self.device.get_info(cl.device_info.PREFERRED_VECTOR_WIDTH_DOUBLE)
            == 0
        ):
            raise Exception(
                "Program uses double-precision floats, but this is not supported on chosen device: %s"
                % self.device.name
            )


def apply_size_heuristics(self, size_heuristics, sizes):
    for platform_name, device_type, size, valuef in size_heuristics:
        if (
            sizes[size] == None
            and self.platform.name.find(platform_name) >= 0
            and (self.device.type & device_type) == device_type
        ):
            sizes[size] = valuef(self.device)
    return sizes


def to_c_str_rep(x):
    if type(x) is bool or type(x) is np.bool_:
        if x:
            return "true"
        else:
            return "false"
    else:
        return str(x)


def initialise_opencl_object(
    self,
    program_src="",
    build_options=[],
    command_queue=None,
    interactive=False,
    platform_pref=None,
    device_pref=None,
    default_group_size=None,
    default_num_groups=None,
    default_tile_size=None,
    default_reg_tile_size=None,
    default_threshold=None,
    size_heuristics=[],
    required_types=[],
    all_sizes={},
    user_sizes={},
    constants=[],
):
    if command_queue is None:
        self.ctx = get_prefered_context(
            interactive, platform_pref, device_pref
        )
        self.queue = cl.CommandQueue(self.ctx)
    else:
        self.ctx = command_queue.context
        self.queue = command_queue
    self.device = self.queue.device
    self.platform = self.device.platform
    self.pool = cl.tools.MemoryPool(cl.tools.ImmediateAllocator(self.queue))
    device_type = self.device.type

    check_types(self, required_types)

    max_group_size = int(self.device.max_work_group_size)
    max_tile_size = int(np.sqrt(self.device.max_work_group_size))

    self.max_thread_block_size = max_group_size
    self.max_tile_size = max_tile_size
    self.max_threshold = 0
    self.max_grid_size = 0

    self.max_shared_memory = int(self.device.local_mem_size)

    # Futhark reserves 4 bytes of local memory for its own purposes.
    self.max_shared_memory -= 4

    # See comment in rts/c/opencl.h.
    if self.platform.name.find("NVIDIA CUDA") >= 0:
        self.max_shared_memory -= 12
    elif self.platform.name.find("AMD") >= 0:
        self.max_shared_memory -= 16

    self.max_registers = int(2**16)  # Not sure how to query for this.

    self.max_cache = self.device.get_info(cl.device_info.GLOBAL_MEM_CACHE_SIZE)

    if self.max_cache == 0:
        self.max_cache = 1024 * 1024

    self.free_list = {}

    self.global_failure = self.pool.allocate(np.int32().itemsize)
    cl.enqueue_fill_buffer(
        self.queue, self.global_failure, np.int32(-1), 0, np.int32().itemsize
    )
    self.global_failure_args = self.pool.allocate(
        np.int64().itemsize * (self.global_failure_args_max + 1)
    )
    self.failure_is_an_option = np.int32(0)

    if "default_group_size" in sizes:
        default_group_size = sizes["default_group_size"]
        del sizes["default_group_size"]

    if "default_num_groups" in sizes:
        default_num_groups = sizes["default_num_groups"]
        del sizes["default_num_groups"]

    if "default_tile_size" in sizes:
        default_tile_size = sizes["default_tile_size"]
        del sizes["default_tile_size"]

    if "default_reg_tile_size" in sizes:
        default_reg_tile_size = sizes["default_reg_tile_size"]
        del sizes["default_reg_tile_size"]

    if "default_threshold" in sizes:
        default_threshold = sizes["default_threshold"]
        del sizes["default_threshold"]

    default_group_size_set = default_group_size != None
    default_tile_size_set = default_tile_size != None
    default_sizes = apply_size_heuristics(
        self,
        size_heuristics,
        {
            "group_size": default_group_size,
            "tile_size": default_tile_size,
            "reg_tile_size": default_reg_tile_size,
            "num_groups": default_num_groups,
            "lockstep_width": None,
            "threshold": default_threshold,
        },
    )
    default_group_size = default_sizes["group_size"]
    default_num_groups = default_sizes["num_groups"]
    default_threshold = default_sizes["threshold"]
    default_tile_size = default_sizes["tile_size"]
    default_reg_tile_size = default_sizes["reg_tile_size"]
    lockstep_width = default_sizes["lockstep_width"]

    if default_group_size > max_group_size:
        if default_group_size_set:
            sys.stderr.write(
                "Note: Device limits group size to {} (down from {})\n".format(
                    max_tile_size, default_group_size
                )
            )
        default_group_size = max_group_size

    if default_tile_size > max_tile_size:
        if default_tile_size_set:
            sys.stderr.write(
                "Note: Device limits tile size to {} (down from {})\n".format(
                    max_tile_size, default_tile_size
                )
            )
        default_tile_size = max_tile_size

    for k, v in user_sizes.items():
        if k in all_sizes:
            all_sizes[k]["value"] = v
        else:
            raise Exception(
                "Unknown size: {}\nKnown sizes: {}".format(
                    k, " ".join(all_sizes.keys())
                )
            )

    self.sizes = {}
    for k, v in all_sizes.items():
        if v["class"] == "thread_block_size":
            max_value = max_group_size
            default_value = default_group_size
        elif v["class"] == "grid_size":
            max_value = max_group_size  # Intentional!
            default_value = default_num_groups
        elif v["class"] == "tile_size":
            max_value = max_tile_size
            default_value = default_tile_size
        elif v["class"] == "reg_tile_size":
            max_value = None
            default_value = default_reg_tile_size
        elif v["class"].startswith("shared_memory"):
            max_value = self.max_shared_memory
            default_value = self.max_shared_memory
        elif v["class"].startswith("cache"):
            max_value = self.max_cache
            default_value = self.max_cache
        elif v["class"].startswith("threshold"):
            max_value = None
            default_value = default_threshold
        else:
            # Bespoke sizes have no limit or default.
            max_value = None
        if v["value"] == None:
            self.sizes[k] = default_value
        elif max_value != None and v["value"] > max_value:
            sys.stderr.write(
                "Note: Device limits {} to {} (down from {}\n".format(
                    k, max_value, v["value"]
                )
            )
            self.sizes[k] = max_value
        else:
            self.sizes[k] = v["value"]

    # XXX: we perform only a subset of z-encoding here.  Really, the
    # compiler should provide us with the variables to which
    # parameters are mapped.
    if len(program_src) >= 0:
        build_options += ["-DLOCKSTEP_WIDTH={}".format(lockstep_width)]

        build_options += [
            "-D{}={}".format("max_thread_block_size", max_group_size)
        ]

        build_options += [
            "-D{}={}".format(
                s.replace("z", "zz")
                .replace(".", "zi")
                .replace("#", "zh")
                .replace("'", "zq"),
                v,
            )
            for (s, v) in self.sizes.items()
        ]

        build_options += [
            "-D{}={}".format(s, to_c_str_rep(f())) for (s, f) in constants
        ]

        if self.platform.name == "Oclgrind":
            build_options += ["-DEMULATE_F16"]

        build_options += [
            f"-DTR_BLOCK_DIM={TR_BLOCK_DIM}",
            f"-DTR_TILE_DIM={TR_TILE_DIM}",
            f"-DTR_ELEMS_PER_THREAD={TR_ELEMS_PER_THREAD}",
        ]

        program = cl.Program(self.ctx, program_src).build(build_options)

        self.transpose_kernels = {
            1: {
                "default": program.map_transpose_1b,
                "low_height": program.map_transpose_1b_low_height,
                "low_width": program.map_transpose_1b_low_width,
                "small": program.map_transpose_1b_small,
                "large": program.map_transpose_1b_large,
            },
            2: {
                "default": program.map_transpose_2b,
                "low_height": program.map_transpose_2b_low_height,
                "low_width": program.map_transpose_2b_low_width,
                "small": program.map_transpose_2b_small,
                "large": program.map_transpose_2b_large,
            },
            4: {
                "default": program.map_transpose_4b,
                "low_height": program.map_transpose_4b_low_height,
                "low_width": program.map_transpose_4b_low_width,
                "small": program.map_transpose_4b_small,
                "large": program.map_transpose_4b_large,
            },
            8: {
                "default": program.map_transpose_8b,
                "low_height": program.map_transpose_8b_low_height,
                "low_width": program.map_transpose_8b_low_width,
                "small": program.map_transpose_8b_small,
                "large": program.map_transpose_8b_large,
            },
        }

        self.copy_kernels = {
            1: program.lmad_copy_1b,
            2: program.lmad_copy_2b,
            4: program.lmad_copy_4b,
            8: program.lmad_copy_8b,
        }

        return program


def opencl_alloc(self, min_size, tag):
    min_size = 1 if min_size == 0 else min_size
    assert min_size > 0
    return self.pool.allocate(min_size)


def opencl_free_all(self):
    self.pool.free_held()


def sync(self):
    failure = np.empty(1, dtype=np.int32)
    cl.enqueue_copy(self.queue, failure, self.global_failure, is_blocking=True)
    self.failure_is_an_option = np.int32(0)
    if failure[0] >= 0:
        # Reset failure information.
        cl.enqueue_fill_buffer(
            self.queue,
            self.global_failure,
            np.int32(-1),
            0,
            np.int32().itemsize,
        )

        # Read failure args.
        failure_args = np.empty(
            self.global_failure_args_max + 1, dtype=np.int64
        )
        cl.enqueue_copy(
            self.queue,
            failure_args,
            self.global_failure_args,
            is_blocking=True,
        )

        raise Exception(self.failure_msgs[failure[0]].format(*failure_args))


def map_transpose_gpu2gpu(
    self, elem_size, dst, dst_offset, src, src_offset, k, n, m
):
    kernels = self.transpose_kernels[elem_size]
    kernel = kernels["default"]
    mulx = TR_BLOCK_DIM / n
    muly = TR_BLOCK_DIM / m

    group_dims = (TR_TILE_DIM, TR_TILE_DIM // TR_ELEMS_PER_THREAD, 1)
    dims = (
        (m + TR_TILE_DIM - 1) // TR_TILE_DIM * group_dims[0],
        (n + TR_TILE_DIM - 1) // TR_TILE_DIM * group_dims[1],
        k,
    )

    k32 = np.int32(k)
    n32 = np.int32(n)
    m32 = np.int32(m)
    mulx32 = np.int32(mulx)
    muly32 = np.int32(muly)

    kernel.set_args(
        cl.LocalMemory(TR_TILE_DIM * (TR_TILE_DIM + 1) * elem_size),
        dst,
        dst_offset,
        src,
        src_offset,
        k32,
        m32,
        n32,
        mulx32,
        muly32,
        np.int32(0),
        np.int32(0),
    )
    cl.enqueue_nd_range_kernel(self.queue, kernel, dims, group_dims)


def copy_elements_gpu2gpu(
    self,
    elem_size,
    dst,
    dst_offset,
    dst_strides,
    src,
    src_offset,
    src_strides,
    shape,
):
    r = len(shape)
    if r > 8:
        raise Exception(
            "Futhark runtime limitation:\nCannot copy array of greater than rank 8.\n"
        )

    n = np.prod(shape)
    zero = np.int64(0)
    layout_args = [None] * (8 * 3)
    for i in range(8):
        if i < r:
            layout_args[i * 3 + 0] = shape[i]
            layout_args[i * 3 + 1] = dst_strides[i]
            layout_args[i * 3 + 2] = src_strides[i]
        else:
            layout_args[i * 3 + 0] = zero
            layout_args[i * 3 + 1] = zero
            layout_args[i * 3 + 2] = zero

    kernel = self.copy_kernels[elem_size]
    kernel.set_args(
        cl.LocalMemory(1),
        dst,
        dst_offset,
        src,
        src_offset,
        n,
        np.int32(r),
        *layout_args,
    )
    w = 256
    dims = ((n + w - 1) // w * w,)
    group_dims = (w,)
    cl.enqueue_nd_range_kernel(self.queue, kernel, dims, group_dims)


def lmad_copy_gpu2gpu(
    self, pt, dst, dst_offset, dst_strides, src, src_offset, src_strides, shape
):
    elem_size = ct.sizeof(pt)
    nbytes = np.prod(shape) * elem_size
    if nbytes == 0:
        return None
    if lmad_memcpyable(dst_strides, src_strides, shape):
        cl.enqueue_copy(
            self.queue,
            dst,
            src,
            dst_offset=dst_offset * elem_size,
            src_offset=src_offset * elem_size,
            byte_count=nbytes,
        )
    else:
        tr = lmad_map_tr(dst_strides, src_strides, shape)
        if tr is not None:
            (k, n, m) = tr
            map_transpose_gpu2gpu(
                self, elem_size, dst, dst_offset, src, src_offset, k, m, n
            )
        else:
            copy_elements_gpu2gpu(
                self,
                elem_size,
                dst,
                dst_offset,
                dst_strides,
                src,
                src_offset,
                src_strides,
                shape,
            )



#!/usr/bin/env python3
#
# Turn a PNG image into a Futhark value of type [height][width]i32 encoded in
# the binary data format.
# Usage: ./png2data image.png output.data
#
# Absolutely no error checking is done.
#
# png2data currently supports 24-bit RGB PNGs. Transparency, grayscale,
# multiple palettes and most other features of PNG is not supported.

import sys
import numpy as np
import png

if __name__ == "__main__":
    infile = sys.argv[1]
    outfile = sys.argv[2]

    r = png.Reader(infile)
    (width, height, img, _) = r.read()
    image_2d = np.vstack(list(map(np.uint32, img)))
    image_3d = np.reshape(image_2d, (height, width, 3))
    array = np.empty((height, width), dtype=np.int32)
    array = array | (image_3d[:, :, 0] << 16)
    array = array | (image_3d[:, :, 1] << 8)
    array = array | (image_3d[:, :, 2])

    with open(outfile, "wb") as f:
        f.write(b"b")
        f.write(np.int8(2))  # type: ignore
        f.write(np.int8(2))  # type: ignore
        f.write(b" i32")
        f.write(np.uint64(height))  # type: ignore
        f.write(np.uint64(width))  # type: ignore
        array.tofile(f)



#!/usr/bin/env python3

from cycler import cycler
import sys
import argparse
import matplotlib.pyplot as plt  # type: ignore
import matplotlib
import numpy as np
import json
import os
import pathlib
import textwrap
import re
import time
import random
import string
from matplotlib.ticker import FormatStrFormatter
from multiprocessing import Pool
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, List, Tuple, NamedTuple, Set
from itertools import islice
from PIL import ImageFile
from collections import OrderedDict

assert sys.version_info >= (3, 9), "Use Python 3.9 or newer."
ImageFile.LOAD_TRUNCATED_IMAGES = True


CSS = """
h1 { font-size: 30px; }
h2 { font-size: 22px; }
h3 { font-size: 18px; }
h1, h2, h3 {
  background-color: #5f021f;
  color: #ffffff;
}
h3 { width: 50%; }
h1 a, h1 a:visited, h2 a, h2 a:visited, h3 a, h3 a:visited {
  color: #fff9e5;
  text-decoration: none;
}
img { width: 25% }
body {
    font-family: sans-serif;
    padding: 0px;
    margin: 0px;
    margin-left: auto;
    margin-right: auto;
    overflow-y: scroll;
    line-height: 1.7;
}
section        { border: 2px solid transparent; }
section:target { border: 2px solid black; }
"""


class PlotJob(NamedTuple):
    program_path: str
    program: str
    dataset: str
    benchmark_result: Dict[str, Any]
    plots: Dict[str, str]


def mpe(runtimes: Optional[np.ndarray[Any, Any]] = None, **kwargs) -> str:
    """Computes the Mean percentage error and formats for printing."""

    if runtimes is None:
        raise Exception(f"runtimes has to be not None.")

    factor = 100 / runtimes.shape[0]
    mpe = factor * ((runtimes - runtimes.mean()) / runtimes).sum()
    return f"{mpe:.5f}%"


def memory_usage(bytes: Optional[Dict[str, int]] = None, **kwargs) -> str:
    """Computes the memory usages of devices and formats for printing."""

    if bytes is None:
        raise Exception(f"bytes has to be not None.")

    def formatter(device: str, bs: int) -> str:
        return f"{format_bytes(bs)}@{device}"

    return ", ".join(map(lambda a: formatter(*a), bytes.items()))


def confidence_interval(
    runtimes: Optional[np.ndarray[Any, Any]] = None, **kwargs
) -> str:
    """Computes the 95% confidence interval and formats for printing."""

    if runtimes is None:
        raise Exception(f"runtimes has to be not None.")

    mean = runtimes.mean()
    bound = 0.95 * runtimes.std(ddof=1) / np.sqrt(runtimes.shape[0])
    return f"[{format_time(mean - bound)}; {format_time(mean + bound)}]"


# Here other descriptors can be added.
DESCRIPTORS = {
    "Mean Percentage Error": mpe,
    "Memory Usage": memory_usage,
    "95% Confidence Interval": confidence_interval,
}


def random_string(size: int) -> str:
    """Creates a random alphanumeric string of a given size."""
    letters = string.ascii_letters + string.digits
    return "".join(random.choice(letters) for _ in range(size))


def format_bytes(x: int) -> str:
    """Tries to find a suitable unit for input x given in bytes."""

    units = [
        ("TiB", 1 / (1024**4)),
        ("GiB", 1 / (1024**3)),
        ("MiB", 1 / (1024**2)),
        ("KiB", 1 / 1024),
        ("bytes", 1),
    ]

    for unit, factor in units:
        temp = factor * x
        if temp > 1:
            return f"{temp:.2f}{unit}"

    return f"{x * units[-1][1]:.2f}{units[-1][0]}"


def format_time(x: int) -> str:
    """Tries to find a suitable time unit for input x."""

    units = [
        ("h", 1 / (60 * 60 * 10**6)),
        ("min", 1 / (60 * 10**6)),
        ("s", 10 ** (-6)),
        ("ms", 10 ** (-3)),
        ("µs", 1),
    ]

    for unit, factor in units:
        temp = factor * x
        if temp > 1:
            return f"{temp:.2f}{unit}"

    return f"{x * units[-1][1]:.2f}{units[-1][0]}"


class PlotType(ABC):
    @abstractmethod
    def plot(self, ax, **kwargs) -> None:
        """Method used to create a plot."""
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def name(cls) -> str:
        """The name of the plot."""
        raise NotImplementedError()

    def __eq__(self, other) -> bool:
        return self.name() == other.name()

    def __lt__(self, other) -> bool:
        return self.name() < other.name()

    @classmethod
    def find_str_formatter(cls, z) -> Tuple[FormatStrFormatter, float]:
        """
        Tries to find a suitable time unit for the given numpy array. The
        scaling factor is returned and a string formatter for matplotlib axis
        using the suitable unit is returned.
        """
        units = [
            (FormatStrFormatter("$%.2fh$"), 1 / (60 * 60 * 10**6)),
            (FormatStrFormatter("$%.2fmin$"), 1 / (60 * 10**6)),
            (FormatStrFormatter("$%.2fs$"), 10 ** (-6)),
            (FormatStrFormatter("$%.2fms$"), 10 ** (-3)),
            (FormatStrFormatter("$%.2f\\mu s$"), 1.0),
        ]

        for unit, factor in units:
            if factor * z.max() > 1:
                return unit, factor

        return units[-1][0], units[-1][1]


PLOT_TYPES_USED: List[str]


class PerRun(PlotType):
    """Create a plot with runtime vs iteration number as plot."""

    def plot(self, ax, runtimes=None, **kwargs) -> None:
        ax.set_title("Per Runtime")
        x = np.arange(len(runtimes))
        y = runtimes
        formatter, factor = PerRun.find_str_formatter(y)
        y = y * factor
        runtimes = ax.scatter(x, y, marker=".")
        ax.legend([runtimes], ["Runtimes"])
        ax.xaxis.set_tick_params(rotation=45)
        ax.yaxis.set_major_formatter(formatter)
        ax.set_ylabel("Runtime")
        ax.set_xlabel("$i$th Runtime")
        ax.grid()

    @classmethod
    def name(cls) -> str:
        return "per_run"


class CumsumPerRun(PlotType):
    """Create a plot with cumulative runtime vs iteration number as plot."""

    def plot(self, ax, runtimes=None, **kwargs) -> None:
        ax.set_title("Cumulative runtime")
        x = np.arange(len(runtimes))
        y = np.cumsum(runtimes)
        formatter, factor = CumsumPerRun.find_str_formatter(y)
        y = y * factor
        X = np.vstack([x, np.ones(len(x))]).T
        slope, intercept = np.linalg.lstsq(X, y, rcond=None)[0]
        ax.scatter(x, y, marker=".")
        ax.xaxis.set_tick_params(rotation=45)
        ax.plot(x, slope * x + intercept, color="black")
        ax.set_ylabel("Cumulative Runtime")
        ax.set_xlabel("$i$th Runtime")
        ax.yaxis.set_major_formatter(formatter)
        ax.grid()

    @classmethod
    def name(cls) -> str:
        return "cumsum_per_run"


class RuntimeDensities(PlotType):
    """Creates a plots the probability density of the runtimes."""

    def plot(self, ax, runtimes=None, **kwargs) -> None:
        ax.set_title("Runtime Densities")
        bincount = np.trim_zeros(np.bincount(runtimes))
        y = bincount / len(runtimes)
        x = np.arange(runtimes.min(), runtimes.max() + 1)
        formatter, factor = RuntimeDensities.find_str_formatter(x)
        x = x * factor
        ax.xaxis.set_tick_params(rotation=45)
        ax.xaxis.set_major_formatter(formatter)
        mean = ax.axvline(x=runtimes.mean() * factor, color="k", label="mean")
        ymin = y.min()
        ymax = y.max()
        padding = abs(ymax - ymin) * 0.05
        ax.set_ylim(ymin - padding, ymax + padding)
        ax.legend([mean], ["Mean Runtime"])
        ax.set_xlabel("Runtime")
        ax.set_ylabel("Density")
        ax.plot(x, y, linestyle="-")
        ax.grid()

    @classmethod
    def name(cls) -> str:
        return "runtime_densities"


class LagPlot(PlotType):
    """
    Creates a lag plot where given some runtimes it copies the array and
    shifts them by one and then plots the two data points vs each other.
    """

    def plot(self, ax, runtimes=None, **kwargs) -> None:
        ax.set_title("Lag Plot")
        x = runtimes
        formatter, factor = LagPlot.find_str_formatter(x)
        x = x * factor
        y = np.roll(x, 1)
        ax.yaxis.set_major_formatter(formatter)
        ax.xaxis.set_major_formatter(formatter)
        ax.xaxis.set_tick_params(rotation=45)
        ax.set_xlabel("The $i$th Runtime")
        ax.set_ylabel("The $i$th + 1 Runtime")
        ax.scatter(x, y, marker=".")
        ax.grid()

    @classmethod
    def name(cls) -> str:
        return "lag_plot"


ALL_PLOT_TYPES = OrderedDict(
    {plot_type.name(): plot_type for plot_type in PlotType.__subclasses__()}
)


class Plotter:
    """Class that will plot and save many figures on a process."""

    def __init__(
        self,
        plot_types: List[PlotType],
        dpi: Any = "200",
        transparent: bool = False,
    ) -> None:
        self.dpi = dpi
        self.plot_types = list(sorted(plot_types))
        self.fig, self.ax = plt.subplots(figsize=(6.4, 5.8))
        self.transparent = transparent
        self.backends = {
            ".png": "AGG",
            ".pdf": "PDF",
            ".ps": "PS",
            ".eps": "PS",
            ".svg": "SVG",
            ".pgf": "PGF",
        }

    def plot(self, data: Dict[str, PlotJob]) -> None:
        """
        Will use the plotter function on all the given data. The data is a
        dictionary where the key is the destination and the value is the values
        that will be passed to the plotter function.
        """

        for datapoint in data.values():
            plots = datapoint.plots
            for plotter in self.plot_types:
                plotter.plot(self.ax, **datapoint.benchmark_result)

                program = datapoint.program
                dataset = datapoint.dataset
                dest = plots[plotter.name()]

                suptitle = textwrap.shorten(rf"{program}: {dataset}", 50)
                self.fig.suptitle(suptitle)
                self.fig.tight_layout()
                ext = pathlib.Path(dest).suffix
                self.fig.savefig(
                    dest,
                    bbox_inches="tight",
                    dpi=self.dpi,
                    backend=self.backends[ext],
                    transparent=self.transparent,
                )
                print(dest)
                self.ax.cla()
        plt.close(self.fig)


def chunks(data: Dict, size: int):
    """Generator that makes sub-dictionaries of a maximum size."""
    it = iter(data)
    for _ in range(0, len(data), size):
        yield {k: data[k] for k in islice(it, size)}


def get_args() -> Any:
    """Gets the arguments used in the program."""
    parser = argparse.ArgumentParser(
        prog="Futhark Plots", description="Makes plots for futhark benchmarks."
    )
    parser.add_argument(
        "filename",
        metavar="FILE",
        help=(
            "the benchmark results as a json file generated by futhark "
            "bench."
        ),
    )
    parser.add_argument(
        "--programs",
        metavar="PROGRAM0,PROGRAM1,...",
        help=(
            "the specific programs the plots will be generated from. Default"
            "is all programs."
        ),
    )
    parser.add_argument(
        "--plots",
        metavar="PLOTTYPE0,PLOTTYPE1,...",
        help=(
            f"the type of plots that will be generated which can be "
            f'{", ".join(ALL_PLOT_TYPES.keys())}. Default is all plots.'
        ),
    )
    parser.add_argument(
        "--filetype",
        default="png",
        metavar="BACKEND",
        help=(
            "the file type used, these can be found on the matplotlib "
            "website."
        ),
    )
    parser.add_argument(
        "--transparent",
        action="store_true",
        help="flag to use if the bagground should be transparent.",
    )

    return parser.parse_args()


def format_arg_list(args: Optional[str]) -> Optional[Set[str]]:
    """
    Takes a string of form 'a, b, c, d' and makes a list ['a', 'b', 'c', 'd']
    """
    if args is None:
        return None

    return set(map(lambda arg: arg.strip(), args.split(",")))


def make_plot_jobs_and_directories(
    programs: List[str],
    data: Dict[str, Dict[str, Dict[str, Any]]],
    file_type: str,
    plot_types: List[str],
    root: str = "graphs",
) -> Tuple[Dict[str, PlotJob], Dict[str, List[str]]]:
    """Makes dictionary with plot jobs where plot_jobs are the jobs."""

    plot_jobs = dict()
    folder_content: Dict[str, List[str]] = dict()

    def remove_characters(characters: List[str], text: str) -> str:
        rep = {re.escape(k): "" for k in characters}
        pattern = re.compile("|".join(rep.keys()))
        return pattern.sub(lambda m: rep[re.escape(m.group(0))], text)

    for program_path in programs:
        temp = data.get(program_path)
        if temp is None:
            raise Exception(f"{program_path} is not valid key.")

        datasets = temp.get("datasets")

        if datasets is None:
            raise Exception(f"{program_path} does not have a dataset key.")

        program_name = pathlib.Path(program_path).name
        program_directory = os.path.dirname(program_path)
        for dataset_path, dataset_dict in datasets.items():
            dataset_name = pathlib.Path(dataset_path).name
            bad_chars = [" ", "#", '"', "/"]
            dataset_path = remove_characters(bad_chars, dataset_path)
            dataset_name_striped = dataset_path.replace(".in", "")
            raw_filename = f"{program_name}_{dataset_name_striped}"
            dataset_filename = raw_filename[:100].replace(" ", "_")
            directory = os.path.join(
                root, program_directory, pathlib.Path(program_path).name
            )
            directory = "." if directory == "" else directory

            benchmark_result = dataset_dict.copy()
            np_runtimes = np.array(benchmark_result.get("runtimes"))
            benchmark_result["runtimes"] = np_runtimes

            os.makedirs(directory, exist_ok=True)

            if folder_content.get(directory) is None:
                folder_content[directory] = []

            while True:
                plot_file_name = os.path.join(
                    directory, f"{dataset_filename}_{random_string(16)}"
                )
                if plot_file_name not in folder_content[directory]:
                    break

            folder_content[directory].insert(0, plot_file_name)

            plot_jobs[plot_file_name] = PlotJob(
                program_path,
                program_name,
                dataset_name,
                benchmark_result,
                {
                    plot_type: f"{plot_file_name}_{plot_type}.{file_type}"
                    for plot_type in plot_types
                },
            )

    return plot_jobs, folder_content


def make_html_descriptors(plot_job: PlotJob) -> str:
    """Makes a table with statistical descriptors for the plot_job"""

    def row(name, func):
        result = func(**plot_job.benchmark_result)
        return rf"<tr><td>{name}:</td><td>{result}</td></tr>"

    return f"""<table>
    <tbody>
        {''.join(map(lambda a: row(*a), DESCRIPTORS.items()))}
    </tbody>
</table>"""


def make_html(
    folder_content: Dict[str, List[str]],
    plot_jobs: Dict[str, PlotJob],
    root: str,
) -> str:
    """Makes a simpel html document with links to each section with plots."""

    def make_key(s, size):
        return f'{"".join(e for e in s if e.isalnum())}{random_string(size)}'

    plot_jobs_keys = dict()
    for key in plot_jobs.keys():
        program = plot_jobs[key].program
        dataset = plot_jobs[key].dataset

        while True:
            id_key = make_key(program + dataset, 32)
            if id_key not in plot_jobs.values():
                break

        plot_jobs_keys[key] = id_key

    folder_keys: Dict[str, str] = dict()
    for folder in folder_content:
        while True:
            id_folder_key = make_key(folder, 32)
            if id_folder_key not in folder_keys.values():
                break

        folder_keys[folder] = id_folder_key

    root_prefix = f"{root}/"

    def make_li(p: str) -> str:
        """
        Makes a single bullet point for a given benchmark's dataset.
        """
        dataset = plot_jobs[p].dataset
        key = plot_jobs_keys[p]
        return rf"<li><a href=#{key}>{dataset}</a></li>"

    def make_list(path: str) -> str:
        """
        Creates the list which shows the structure of benchmarks and links to
        the sections.
        """
        sorted_paths = sorted(os.listdir(path))
        directory = list(map(lambda p: os.path.join(path, p), sorted_paths))

        isdir = os.path.isdir

        if len(directory) == 1 and isdir(directory[0]) and path == root:
            return make_list(directory[0])

        before = "".join(map(make_li, folder_content.get(path, [])))
        lis = "".join(map(make_list, directory))
        pretty_path = path.removeprefix(root_prefix)

        if folder_keys.get(path) is not None:
            pretty_path = f"<a href=#{folder_keys.get(path)}>{pretty_path}</a>"

        return rf"<li>{pretty_path}</li><ul>{before}{lis}</ul>"

    def make_subsection(plot_file: str, plot_job: Optional[PlotJob]) -> str:
        """
        Makes a subsection with plots and statistical descriptors.
        """
        if plot_job is None:
            raise Exception(f"A given PlotJob was None.")

        dataset = plot_job.dataset
        program = plot_job.program
        key = plot_jobs_keys[plot_file]
        descriptors = make_html_descriptors(plot_job)

        plots = "".join(
            map(lambda plot: f"<img src='{plot}'/>", plot_job.plots.values())
        )
        header = rf"<h3><a href=#{key}>{program}: {dataset}</a></h3>"
        return rf"<section id={key}>{header}{plots}{descriptors}</section>"

    def make_section(folder: str, dataset_plot_files: List[str]) -> str:
        """
        Makes a section with all the plots and descriptors for a given
        benchmark's datasets.
        """
        sub_data = map(lambda a: (a, plot_jobs.get(a)), dataset_plot_files)
        subsections = "".join(map(lambda a: make_subsection(*a), sub_data))
        pretty_folder = folder.removeprefix(root_prefix)
        folder_key = folder_keys[folder]
        header = rf"<h2><a href=#{folder_key}>{pretty_folder}</a></h2>"
        return rf"<section id={folder_key}>{header}{subsections}</section>"

    width = 100 // len(next(iter(plot_jobs.values())).plots)
    lis = make_list(root)
    sorted_content = sorted(folder_content.items())
    sections = "".join(map(lambda a: make_section(*a), sorted_content))
    return f"""<!doctype html>
    <html lang="en">
    <head>
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
    <title>{root}</title>
    <style type="text/css">
      {CSS}
    </style>
    </head>
    <body>
    <header>
    <h1>{root}</h1>
    <nav>
        <ul>
            {lis}
        </ul>
    </nav>
    </header>
    {sections}
    </body>
    </html>"""


def task(plot_jobs: Dict[str, PlotJob]) -> None:
    """Begins plotting, it is used"""
    global plots
    global PLOT_TYPES_USED
    global TRANSPARENT

    plot_types = [
        plot_type()  # type: ignore
        for key, plot_type in ALL_PLOT_TYPES.items()
        if key in PLOT_TYPES_USED
    ]
    plotter = Plotter(plot_types, dpi=200, transparent=TRANSPARENT)
    plotter.plot(plot_jobs)


TRANSPARENT: bool


def main() -> None:
    global PLOT_TYPES_USED
    global TRANSPARENT
    plt.rcParams.update(
        {
            "ytick.color": "black",
            "xtick.color": "black",
            "axes.labelcolor": "black",
            "axes.edgecolor": "black",
            "axes.axisbelow": True,
            "text.usetex": False,
            "axes.prop_cycle": cycler(color=["#5f021f"]),
        }
    )

    args = get_args()
    filename = pathlib.Path(args.filename).stem
    data = json.load(open(args.filename, "r"))
    programs = format_arg_list(args.programs)

    plots_used = format_arg_list(args.plots)

    if plots_used is None:
        PLOT_TYPES_USED = list(sorted(ALL_PLOT_TYPES.keys()))
    else:
        PLOT_TYPES_USED = list(sorted(plots_used))
        temp = list(ALL_PLOT_TYPES.keys())
        for plot_type in PLOT_TYPES_USED:
            if plot_type not in temp:
                existing_plot_types = ", ".join(temp)
                raise Exception(
                    (
                        '"{plot_type}" is not a plot type try '
                        f"{existing_plot_types}"
                    )
                )

    filetype = args.filetype
    TRANSPARENT = args.transparent

    root = f"{filename}-plots"

    if os.path.exists(root):
        raise Exception(
            (
                f'The folder "{root}" must be removed before the plots can be '
                "made."
            )
        )

    if programs is None:
        programs = set(data.keys())
    else:
        programs = set(programs)
        keys = set(data.keys())
        if not programs.issubset(keys):
            diff = ", ".join(programs.difference(keys))
            raise Exception(f'"{diff}" are not valid keys.')

    plot_jobs, folder_content = make_plot_jobs_and_directories(
        list(programs), data, filetype, PLOT_TYPES_USED, root=root
    )

    with open(f"{filename}.html", "w") as fp:
        fp.write(make_html(folder_content, plot_jobs, root))

    with Pool(16) as p:
        p.map(task, chunks(plot_jobs, max(len(plot_jobs) // 32, 1)))

    print(f"Open {filename}.html in a browser.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Convert the contents of a file to a Futhark-compatible binary data
# file.  The file is interpreted as raw binary data corresponding to a
# one-dimensional array.
#
# Takes a single option: the name of the element type.  Only the
# numeric types are supported (no floats).
#
# Usage example:
#
# $ ./data2fut.py f32 INPUT OUTPUT

import futhark_data as fd
import numpy as np
import sys

types = {
    "f16": np.float16,
    "f32": np.float32,
    "f64": np.float64,
    "i8": np.int8,
    "i16": np.int16,
    "i32": np.int32,
    "i64": np.int64,
    "u8": np.uint8,
    "u16": np.uint16,
    "u32": np.uint32,
    "u64": np.uint64,
}

_, type, input, output = sys.argv
dtype = types[type]

v = np.fromfile(open(input, mode="rb"), dtype)
fd.dump(v, open(output, mode="wb"), binary=True)



# Start of scalar.py.

import numpy as np
import math
import struct


def intlit(t, x):
    if t == np.int8:
        return np.int8(x)
    elif t == np.int16:
        return np.int16(x)
    elif t == np.int32:
        return np.int32(x)
    else:
        return np.int64(x)


def signed(x):
    if type(x) == np.uint8:
        return np.int8(x)
    elif type(x) == np.uint16:
        return np.int16(x)
    elif type(x) == np.uint32:
        return np.int32(x)
    else:
        return np.int64(x)


def unsigned(x):
    if type(x) == np.int8:
        return np.uint8(x)
    elif type(x) == np.int16:
        return np.uint16(x)
    elif type(x) == np.int32:
        return np.uint32(x)
    else:
        return np.uint64(x)


def shlN(x, y):
    return x << y


def ashrN(x, y):
    return x >> y


# Python is so slow that we just make all the unsafe operations safe,
# always.


def sdivN(x, y):
    if y == 0:
        return intlit(type(x), 0)
    else:
        return x // y


def sdiv_upN(x, y):
    if y == 0:
        return intlit(type(x), 0)
    else:
        return (x + y - intlit(type(x), 1)) // y


def smodN(x, y):
    if y == 0:
        return intlit(type(x), 0)
    else:
        return x % y


def udivN(x, y):
    if y == 0:
        return intlit(type(x), 0)
    else:
        return signed(unsigned(x) // unsigned(y))


def udiv_upN(x, y):
    if y == 0:
        return intlit(type(x), 0)
    else:
        return signed(
            (unsigned(x) + unsigned(y) - unsigned(intlit(type(x), 1)))
            // unsigned(y)
        )


def umodN(x, y):
    if y == 0:
        return intlit(type(x), 0)
    else:
        return signed(unsigned(x) % unsigned(y))


def squotN(x, y):
    if y == 0:
        return intlit(type(x), 0)
    else:
        return np.floor_divide(np.abs(x), np.abs(y)) * np.sign(x) * np.sign(y)


def sremN(x, y):
    if y == 0:
        return intlit(type(x), 0)
    else:
        return np.remainder(np.abs(x), np.abs(y)) * np.sign(x)


def sminN(x, y):
    return min(x, y)


def smaxN(x, y):
    return max(x, y)


def uminN(x, y):
    return signed(min(unsigned(x), unsigned(y)))


def umaxN(x, y):
    return signed(max(unsigned(x), unsigned(y)))


def fminN(x, y):
    return np.fmin(x, y)


def fmaxN(x, y):
    return np.fmax(x, y)


def powN(x, y):
    return x**y


def fpowN(x, y):
    return x**y


def sleN(x, y):
    return x <= y


def sltN(x, y):
    return x < y


def uleN(x, y):
    return unsigned(x) <= unsigned(y)


def ultN(x, y):
    return unsigned(x) < unsigned(y)


def lshr8(x, y):
    return np.int8(np.uint8(x) >> np.uint8(y))


def lshr16(x, y):
    return np.int16(np.uint16(x) >> np.uint16(y))


def lshr32(x, y):
    return np.int32(np.uint32(x) >> np.uint32(y))


def lshr64(x, y):
    return np.int64(np.uint64(x) >> np.uint64(y))


def sext_T_i8(x):
    return np.int8(x)


def sext_T_i16(x):
    return np.int16(x)


def sext_T_i32(x):
    return np.int32(x)


def sext_T_i64(x):
    return np.int64(x)


def itob_T_bool(x):
    return bool(x)


def btoi_bool_i8(x):
    return np.int8(x)


def btoi_bool_i16(x):
    return np.int16(x)


def btoi_bool_i32(x):
    return np.int32(x)


def btoi_bool_i64(x):
    return np.int64(x)


def ftob_T_bool(x):
    return bool(x)


def btof_bool_f16(x):
    return np.float16(x)


def btof_bool_f32(x):
    return np.float32(x)


def btof_bool_f64(x):
    return np.float64(x)


def zext_i8_i8(x):
    return np.int8(np.uint8(x))


def zext_i8_i16(x):
    return np.int16(np.uint8(x))


def zext_i8_i32(x):
    return np.int32(np.uint8(x))


def zext_i8_i64(x):
    return np.int64(np.uint8(x))


def zext_i16_i8(x):
    return np.int8(np.uint16(x))


def zext_i16_i16(x):
    return np.int16(np.uint16(x))


def zext_i16_i32(x):
    return np.int32(np.uint16(x))


def zext_i16_i64(x):
    return np.int64(np.uint16(x))


def zext_i32_i8(x):
    return np.int8(np.uint32(x))


def zext_i32_i16(x):
    return np.int16(np.uint32(x))


def zext_i32_i32(x):
    return np.int32(np.uint32(x))


def zext_i32_i64(x):
    return np.int64(np.uint32(x))


def zext_i64_i8(x):
    return np.int8(np.uint64(x))


def zext_i64_i16(x):
    return np.int16(np.uint64(x))


def zext_i64_i32(x):
    return np.int32(np.uint64(x))


def zext_i64_i64(x):
    return np.int64(np.uint64(x))


sdiv8 = sdiv16 = sdiv32 = sdiv64 = sdivN
sdiv_up8 = sdiv1_up6 = sdiv_up32 = sdiv_up64 = sdiv_upN
sdiv_safe8 = sdiv1_safe6 = sdiv_safe32 = sdiv_safe64 = sdivN
sdiv_up_safe8 = sdiv_up1_safe6 = sdiv_up_safe32 = sdiv_up_safe64 = sdiv_upN
smod8 = smod16 = smod32 = smod64 = smodN
smod_safe8 = smod_safe16 = smod_safe32 = smod_safe64 = smodN
udiv8 = udiv16 = udiv32 = udiv64 = udivN
udiv_up8 = udiv_up16 = udiv_up32 = udiv_up64 = udivN
udiv_safe8 = udiv_safe16 = udiv_safe32 = udiv_safe64 = udiv_upN
udiv_up_safe8 = udiv_up_safe16 = udiv_up_safe32 = udiv_up_safe64 = udiv_upN
umod8 = umod16 = umod32 = umod64 = umodN
umod_safe8 = umod_safe16 = umod_safe32 = umod_safe64 = umodN
squot8 = squot16 = squot32 = squot64 = squotN
squot_safe8 = squot_safe16 = squot_safe32 = squot_safe64 = squotN
srem8 = srem16 = srem32 = srem64 = sremN
srem_safe8 = srem_safe16 = srem_safe32 = srem_safe64 = sremN

shl8 = shl16 = shl32 = shl64 = shlN
ashr8 = ashr16 = ashr32 = ashr64 = ashrN
smax8 = smax16 = smax32 = smax64 = smaxN
smin8 = smin16 = smin32 = smin64 = sminN
umax8 = umax16 = umax32 = umax64 = umaxN
umin8 = umin16 = umin32 = umin64 = uminN
pow8 = pow16 = pow32 = pow64 = powN
fpow16 = fpow32 = fpow64 = fpowN
fmax16 = fmax32 = fmax64 = fmaxN
fmin16 = fmin32 = fmin64 = fminN
sle8 = sle16 = sle32 = sle64 = sleN
slt8 = slt16 = slt32 = slt64 = sltN
ule8 = ule16 = ule32 = ule64 = uleN
ult8 = ult16 = ult32 = ult64 = ultN
sext_i8_i8 = sext_i16_i8 = sext_i32_i8 = sext_i64_i8 = sext_T_i8
sext_i8_i16 = sext_i16_i16 = sext_i32_i16 = sext_i64_i16 = sext_T_i16
sext_i8_i32 = sext_i16_i32 = sext_i32_i32 = sext_i64_i32 = sext_T_i32
sext_i8_i64 = sext_i16_i64 = sext_i32_i64 = sext_i64_i64 = sext_T_i64
itob_i8_bool = itob_i16_bool = itob_i32_bool = itob_i64_bool = itob_T_bool
ftob_f16_bool = ftob_f32_bool = ftob_f64_bool = ftob_T_bool


def clz_T(x):
    n = np.int32(0)
    bits = x.itemsize * 8
    for i in range(bits):
        if x < 0:
            break
        n += np.int32(1)
        x <<= np.int8(1)
    return n


def ctz_T(x):
    n = np.int32(0)
    bits = x.itemsize * 8
    for i in range(bits):
        if (x & 1) == 1:
            break
        n += np.int32(1)
        x >>= np.int8(1)
    return n


def popc_T(x):
    c = np.int32(0)
    while x != 0:
        x &= x - np.int8(1)
        c += np.int32(1)
    return c


futhark_popc8 = futhark_popc16 = futhark_popc32 = futhark_popc64 = popc_T
futhark_clzz8 = futhark_clzz16 = futhark_clzz32 = futhark_clzz64 = clz_T
futhark_ctzz8 = futhark_ctzz16 = futhark_ctzz32 = futhark_ctzz64 = ctz_T


def ssignum(x):
    return np.sign(x)


def usignum(x):
    if x < 0:
        return ssignum(-x)
    else:
        return ssignum(x)


def sitofp_T_f32(x):
    return np.float32(x)


sitofp_i8_f32 = sitofp_i16_f32 = sitofp_i32_f32 = sitofp_i64_f32 = sitofp_T_f32


def sitofp_T_f64(x):
    return np.float64(x)


sitofp_i8_f64 = sitofp_i16_f64 = sitofp_i32_f64 = sitofp_i64_f64 = sitofp_T_f64


def uitofp_T_f32(x):
    return np.float32(unsigned(x))


uitofp_i8_f32 = uitofp_i16_f32 = uitofp_i32_f32 = uitofp_i64_f32 = uitofp_T_f32


def uitofp_T_f64(x):
    return np.float64(unsigned(x))


uitofp_i8_f64 = uitofp_i16_f64 = uitofp_i32_f64 = uitofp_i64_f64 = uitofp_T_f64


def fptosi_T_i8(x):
    if np.isnan(x) or np.isinf(x):
        return np.int8(0)
    else:
        return np.int8(np.trunc(x))


fptosi_f16_i8 = fptosi_f32_i8 = fptosi_f64_i8 = fptosi_T_i8


def fptosi_T_i16(x):
    if np.isnan(x) or np.isinf(x):
        return np.int16(0)
    else:
        return np.int16(np.trunc(x))


fptosi_f16_i16 = fptosi_f32_i16 = fptosi_f64_i16 = fptosi_T_i16


def fptosi_T_i32(x):
    if np.isnan(x) or np.isinf(x):
        return np.int32(0)
    else:
        return np.int32(np.trunc(x))


fptosi_f16_i32 = fptosi_f32_i32 = fptosi_f64_i32 = fptosi_T_i32


def fptosi_T_i64(x):
    if np.isnan(x) or np.isinf(x):
        return np.int64(0)
    else:
        return np.int64(np.trunc(x))


fptosi_f16_i64 = fptosi_f32_i64 = fptosi_f64_i64 = fptosi_T_i64


def fptoui_T_i8(x):
    if np.isnan(x) or np.isinf(x):
        return np.int8(0)
    else:
        return np.int8(np.trunc(x))


fptoui_f16_i8 = fptoui_f32_i8 = fptoui_f64_i8 = fptoui_T_i8


def fptoui_T_i16(x):
    if np.isnan(x) or np.isinf(x):
        return np.int16(0)
    else:
        return np.int16(np.trunc(x))


fptoui_f16_i16 = fptoui_f32_i16 = fptoui_f64_i16 = fptoui_T_i16


def fptoui_T_i32(x):
    if np.isnan(x) or np.isinf(x):
        return np.int32(0)
    else:
        return np.int32(np.trunc(x))


fptoui_f16_i32 = fptoui_f32_i32 = fptoui_f64_i32 = fptoui_T_i32


def fptoui_T_i64(x):
    if np.isnan(x) or np.isinf(x):
        return np.int64(0)
    else:
        return np.int64(np.trunc(x))


fptoui_f16_i64 = fptoui_f32_i64 = fptoui_f64_i64 = fptoui_T_i64


def fpconv_f16_f32(x):
    return np.float32(x)


def fpconv_f16_f64(x):
    return np.float64(x)


def fpconv_f32_f16(x):
    return np.float16(x)


def fpconv_f32_f64(x):
    return np.float64(x)


def fpconv_f64_f16(x):
    return np.float16(x)


def fpconv_f64_f32(x):
    return np.float32(x)


def futhark_umul_hi8(a, b):
    return np.int8(
        (np.uint64(np.uint8(a)) * np.uint64(np.uint8(b))) >> np.uint64(8)
    )


def futhark_umul_hi16(a, b):
    return np.int16(
        (np.uint64(np.uint16(a)) * np.uint64(np.uint16(b))) >> np.uint64(16)
    )


def futhark_umul_hi32(a, b):
    return np.int32(
        (np.uint64(np.uint32(a)) * np.uint64(np.uint32(b))) >> np.uint64(32)
    )


def futhark_umul_hi64(a, b):
    return np.int64(np.uint64(int(np.uint64(a)) * int(np.uint64(b)) >> 64))


def futhark_smul_hi8(a, b):
    return np.int8((np.int64(a) * np.int64(b)) >> np.int64(8))


def futhark_smul_hi16(a, b):
    return np.int16((np.int64(a) * np.int64(b)) >> np.int64(16))


def futhark_smul_hi32(a, b):
    return np.int32((np.int64(a) * np.int64(b)) >> np.int64(32))


def futhark_smul_hi64(a, b):
    return np.int64(int(a) * int(b) >> 64)


def futhark_umad_hi8(a, b, c):
    return futhark_umul_hi8(a, b) + c


def futhark_umad_hi16(a, b, c):
    return futhark_umul_hi16(a, b) + c


def futhark_umad_hi32(a, b, c):
    return futhark_umul_hi32(a, b) + c


def futhark_umad_hi64(a, b, c):
    return futhark_umul_hi64(a, b) + c


def futhark_smad_hi8(a, b, c):
    return futhark_smul_hi8(a, b) + c


def futhark_smad_hi16(a, b, c):
    return futhark_smul_hi16(a, b) + c


def futhark_smad_hi32(a, b, c):
    return futhark_smul_hi32(a, b) + c


def futhark_smad_hi64(a, b, c):
    return futhark_smul_hi64(a, b) + c


def futhark_log64(x):
    return np.float64(np.log(x))


def futhark_log2_64(x):
    return np.float64(np.log2(x))


def futhark_log10_64(x):
    return np.float64(np.log10(x))


def futhark_log1p_64(x):
    return np.float64(np.log1p(x))


def futhark_sqrt64(x):
    return np.sqrt(x)


def futhark_cbrt64(x):
    return np.cbrt(x)


def futhark_exp64(x):
    return np.exp(x)


def futhark_cos64(x):
    return np.cos(x)


def futhark_sin64(x):
    return np.sin(x)


def futhark_tan64(x):
    return np.tan(x)


def futhark_acos64(x):
    return np.arccos(x)


def futhark_asin64(x):
    return np.arcsin(x)


def futhark_atan64(x):
    return np.arctan(x)


def futhark_cosh64(x):
    return np.cosh(x)


def futhark_sinh64(x):
    return np.sinh(x)


def futhark_tanh64(x):
    return np.tanh(x)


def futhark_acosh64(x):
    return np.arccosh(x)


def futhark_asinh64(x):
    return np.arcsinh(x)


def futhark_atanh64(x):
    return np.arctanh(x)


def futhark_atan2_64(x, y):
    return np.arctan2(x, y)


def futhark_hypot64(x, y):
    return np.hypot(x, y)


def futhark_gamma64(x):
    return np.float64(math.gamma(x))


def futhark_lgamma64(x):
    return np.float64(math.lgamma(x))


def futhark_erf64(x):
    return np.float64(math.erf(x))


def futhark_erfc64(x):
    return np.float64(math.erfc(x))


def futhark_round64(x):
    return np.round(x)


def futhark_ceil64(x):
    return np.ceil(x)


def futhark_floor64(x):
    return np.floor(x)


def futhark_nextafter64(x, y):
    return np.nextafter(x, y)


def futhark_isnan64(x):
    return np.isnan(x)


def futhark_isinf64(x):
    return np.isinf(x)


def futhark_to_bits64(x):
    s = struct.pack(">d", x)
    return np.int64(struct.unpack(">q", s)[0])


def futhark_from_bits64(x):
    s = struct.pack(">q", x)
    return np.float64(struct.unpack(">d", s)[0])


def futhark_log32(x):
    return np.float32(np.log(x))


def futhark_log2_32(x):
    return np.float32(np.log2(x))


def futhark_log10_32(x):
    return np.float32(np.log10(x))


def futhark_log1p_32(x):
    return np.float32(np.log1p(x))


def futhark_sqrt32(x):
    return np.float32(np.sqrt(x))


def futhark_cbrt32(x):
    return np.float32(np.cbrt(x))


def futhark_exp32(x):
    return np.exp(x)


def futhark_cos32(x):
    return np.cos(x)


def futhark_sin32(x):
    return np.sin(x)


def futhark_tan32(x):
    return np.tan(x)


def futhark_acos32(x):
    return np.arccos(x)


def futhark_asin32(x):
    return np.arcsin(x)


def futhark_atan32(x):
    return np.arctan(x)


def futhark_cosh32(x):
    return np.cosh(x)


def futhark_sinh32(x):
    return np.sinh(x)


def futhark_tanh32(x):
    return np.tanh(x)


def futhark_acosh32(x):
    return np.arccosh(x)


def futhark_asinh32(x):
    return np.arcsinh(x)


def futhark_atanh32(x):
    return np.arctanh(x)


def futhark_atan2_32(x, y):
    return np.arctan2(x, y)


def futhark_hypot32(x, y):
    return np.hypot(x, y)


def futhark_gamma32(x):
    return np.float32(math.gamma(x))


def futhark_lgamma32(x):
    return np.float32(math.lgamma(x))


def futhark_erf32(x):
    return np.float32(math.erf(x))


def futhark_erfc32(x):
    return np.float32(math.erfc(x))


def futhark_round32(x):
    return np.round(x)


def futhark_ceil32(x):
    return np.ceil(x)


def futhark_floor32(x):
    return np.floor(x)


def futhark_nextafter32(x, y):
    return np.nextafter(x, y)


def futhark_isnan32(x):
    return np.isnan(x)


def futhark_isinf32(x):
    return np.isinf(x)


def futhark_to_bits32(x):
    s = struct.pack(">f", x)
    return np.int32(struct.unpack(">l", s)[0])


def futhark_from_bits32(x):
    s = struct.pack(">l", x)
    return np.float32(struct.unpack(">f", s)[0])


def futhark_log16(x):
    return np.float16(np.log(x))


def futhark_log2_16(x):
    return np.float16(np.log2(x))


def futhark_log10_16(x):
    return np.float16(np.log10(x))


def futhark_log1p_16(x):
    return np.float16(np.log1p(x))


def futhark_sqrt16(x):
    return np.float16(np.sqrt(x))


def futhark_cbrt16(x):
    return np.float16(np.cbrt(x))


def futhark_exp16(x):
    return np.exp(x)


def futhark_cos16(x):
    return np.cos(x)


def futhark_sin16(x):
    return np.sin(x)


def futhark_tan16(x):
    return np.tan(x)


def futhark_acos16(x):
    return np.arccos(x)


def futhark_asin16(x):
    return np.arcsin(x)


def futhark_atan16(x):
    return np.arctan(x)


def futhark_cosh16(x):
    return np.cosh(x)


def futhark_sinh16(x):
    return np.sinh(x)


def futhark_tanh16(x):
    return np.tanh(x)


def futhark_acosh16(x):
    return np.arccosh(x)


def futhark_asinh16(x):
    return np.arcsinh(x)


def futhark_atanh16(x):
    return np.arctanh(x)


def futhark_atan2_16(x, y):
    return np.arctan2(x, y)


def futhark_hypot16(x, y):
    return np.hypot(x, y)


def futhark_gamma16(x):
    return np.float16(math.gamma(x))


def futhark_lgamma16(x):
    return np.float16(math.lgamma(x))


def futhark_erf16(x):
    return np.float16(math.erf(x))


def futhark_erfc16(x):
    return np.float16(math.erfc(x))


def futhark_round16(x):
    return np.round(x)


def futhark_ceil16(x):
    return np.ceil(x)


def futhark_floor16(x):
    return np.floor(x)


def futhark_nextafter16(x, y):
    return np.nextafter(x, y)


def futhark_isnan16(x):
    return np.isnan(x)


def futhark_isinf16(x):
    return np.isinf(x)


def futhark_to_bits16(x):
    s = struct.pack(">e", x)
    return np.int16(struct.unpack(">H", s)[0])


def futhark_from_bits16(x):
    s = struct.pack(">H", np.uint16(x))
    return np.float16(struct.unpack(">e", s)[0])


def futhark_lerp16(v0, v1, t):
    return v0 + (v1 - v0) * t


def futhark_lerp32(v0, v1, t):
    return v0 + (v1 - v0) * t


def futhark_lerp64(v0, v1, t):
    return v0 + (v1 - v0) * t


def futhark_ldexp16(x, y):
    return np.ldexp(x, y)


def futhark_ldexp32(x, y):
    return np.ldexp(x, y)


def futhark_ldexp64(x, y):
    return np.ldexp(x, y)


def futhark_mad16(a, b, c):
    return a * b + c


def futhark_mad32(a, b, c):
    return a * b + c


def futhark_mad64(a, b, c):
    return a * b + c


def futhark_fma16(a, b, c):
    return a * b + c


def futhark_fma32(a, b, c):
    return a * b + c


def futhark_fma64(a, b, c):
    return a * b + c


futhark_copysign16 = futhark_copysign32 = futhark_copysign64 = np.copysign


def futhark_cond(x, y, z):
    return y if x else z


futhark_cond_f16 = futhark_cond_f32 = futhark_cond_f64 = futhark_cond
futhark_cond_i18 = futhark_cond_i16 = futhark_cond_i32 = futhark_cond_i64 = (
    futhark_cond
)
futhark_cond_bool = futhark_cond_unit = futhark_cond


# End of scalar.py.



#!/usr/bin/env python
#
# Remove some benchmark from a JSON file produced by futhark-bench's --json
# option.
#
# This is useful if we accidentally added some pointless programs that
# just obscure things.

import json
import sys

remove = sys.argv[1]
files = sys.argv[2:]

removed = False
i = 0
for fp in files:
    with open(fp, "r") as infile:
        try:
            file_json = json.load(infile)
        except Exception as e:
            print("Could not read {}: {}".format(fp, e))
            continue
    for b in file_json.keys():
        if remove in b:
            file_json.pop(b)
            removed = True
    with open(fp, "w") as outfile:
        json.dump(file_json, outfile)
    i += 1
    if i % 100 == 0:
        print("{}/{} files processed...".format(i, len(files)))

if not removed:
    print("Warning: no JSON file contained {}".format(remove))



# Start of values.py.

# Hacky parser/reader/writer for values written in Futhark syntax.
# Used for reading stdin when compiling standalone programs with the
# Python code generator.

import numpy as np
import string
import struct
import sys


class ReaderInput:
    def __init__(self, f):
        self.f = f
        self.lookahead_buffer = []

    def get_char(self):
        if len(self.lookahead_buffer) == 0:
            return self.f.read(1)
        else:
            c = self.lookahead_buffer[0]
            self.lookahead_buffer = self.lookahead_buffer[1:]
            return c

    def unget_char(self, c):
        self.lookahead_buffer = [c] + self.lookahead_buffer

    def get_chars(self, n):
        n1 = min(n, len(self.lookahead_buffer))
        s = b"".join(self.lookahead_buffer[:n1])
        self.lookahead_buffer = self.lookahead_buffer[n1:]
        n2 = n - n1
        if n2 > 0:
            s += self.f.read(n2)
        return s

    def peek_char(self):
        c = self.get_char()
        if c:
            self.unget_char(c)
        return c


def skip_spaces(f):
    c = f.get_char()
    while c != None:
        if c.isspace():
            c = f.get_char()
        elif c == b"-":
            # May be line comment.
            if f.peek_char() == b"-":
                # Yes, line comment. Skip to end of line.
                while c != b"\n" and c != None:
                    c = f.get_char()
            else:
                break
        else:
            break
    if c:
        f.unget_char(c)


def parse_specific_char(f, expected):
    got = f.get_char()
    if got != expected:
        f.unget_char(got)
        raise ValueError
    return True


def parse_specific_string(f, s):
    # This funky mess is intended, and is caused by the fact that if `type(b) ==
    # bytes` then `type(b[0]) == int`, but we need to match each element with a
    # `bytes`, so therefore we make each character an array element
    b = s.encode("utf8")
    bs = [b[i : i + 1] for i in range(len(b))]
    read = []
    try:
        for c in bs:
            parse_specific_char(f, c)
            read.append(c)
        return True
    except ValueError:
        for c in read[::-1]:
            f.unget_char(c)
        raise


def optional(p, *args):
    try:
        return p(*args)
    except ValueError:
        return None


def optional_specific_string(f, s):
    c = f.peek_char()
    # This funky mess is intended, and is caused by the fact that if `type(b) ==
    # bytes` then `type(b[0]) == int`, but we need to match each element with a
    # `bytes`, so therefore we make each character an array element
    b = s.encode("utf8")
    bs = [b[i : i + 1] for i in range(len(b))]
    if c == bs[0]:
        return parse_specific_string(f, s)
    else:
        return False


def sepEndBy(p, sep, *args):
    elems = []
    x = optional(p, *args)
    if x != None:
        elems += [x]
        while optional(sep, *args) != None:
            x = optional(p, *args)
            if x == None:
                break
            else:
                elems += [x]
    return elems


# Assumes '0x' has already been read
def parse_hex_int(f):
    s = b""
    c = f.get_char()
    while c != None:
        if c in b"01234556789ABCDEFabcdef":
            s += c
            c = f.get_char()
        elif c == b"_":
            c = f.get_char()  # skip _
        else:
            f.unget_char(c)
            break
    return str(int(s, 16)).encode("utf8")  # ugh


def parse_int(f):
    s = b""
    c = f.get_char()
    if c == b"0" and f.peek_char() in b"xX":
        c = f.get_char()  # skip X
        return parse_hex_int(f)
    else:
        while c != None:
            if c.isdigit():
                s += c
                c = f.get_char()
            elif c == b"_":
                c = f.get_char()  # skip _
            else:
                f.unget_char(c)
                break
        if len(s) == 0:
            raise ValueError
        return s


def parse_int_signed(f):
    s = b""
    c = f.get_char()

    if c == b"-" and f.peek_char().isdigit():
        return c + parse_int(f)
    else:
        if c != b"+":
            f.unget_char(c)
        return parse_int(f)


def read_str_comma(f):
    skip_spaces(f)
    parse_specific_char(f, b",")
    return b","


def read_str_int(f, s):
    skip_spaces(f)
    x = int(parse_int_signed(f))
    optional_specific_string(f, s)
    return x


def read_str_uint(f, s):
    skip_spaces(f)
    x = int(parse_int(f))
    optional_specific_string(f, s)
    return x


def read_str_i8(f):
    return np.int8(read_str_int(f, "i8"))


def read_str_i16(f):
    return np.int16(read_str_int(f, "i16"))


def read_str_i32(f):
    return np.int32(read_str_int(f, "i32"))


def read_str_i64(f):
    return np.int64(read_str_int(f, "i64"))


def read_str_u8(f):
    return np.uint8(read_str_int(f, "u8"))


def read_str_u16(f):
    return np.uint16(read_str_int(f, "u16"))


def read_str_u32(f):
    return np.uint32(read_str_int(f, "u32"))


def read_str_u64(f):
    return np.uint64(read_str_int(f, "u64"))


def read_char(f):
    skip_spaces(f)
    parse_specific_char(f, b"'")
    c = f.get_char()
    parse_specific_char(f, b"'")
    return c


def read_str_hex_float(f, sign):
    int_part = parse_hex_int(f)
    parse_specific_char(f, b".")
    frac_part = parse_hex_int(f)
    parse_specific_char(f, b"p")
    exponent = parse_int(f)

    int_val = int(int_part, 16)
    frac_val = float(int(frac_part, 16)) / (16 ** len(frac_part))
    exp_val = int(exponent)

    total_val = (int_val + frac_val) * (2.0**exp_val)
    if sign == b"-":
        total_val = -1 * total_val

    return float(total_val)


def read_str_decimal(f):
    skip_spaces(f)
    c = f.get_char()
    if c == b"-":
        sign = b"-"
    else:
        f.unget_char(c)
        sign = b""

    # Check for hexadecimal float
    c = f.get_char()
    if c == "0" and (f.peek_char() in ["x", "X"]):
        f.get_char()
        return read_str_hex_float(f, sign)
    else:
        f.unget_char(c)

    bef = optional(parse_int, f)
    if bef == None:
        bef = b"0"
        parse_specific_char(f, b".")
        aft = parse_int(f)
    elif optional(parse_specific_char, f, b"."):
        aft = parse_int(f)
    else:
        aft = b"0"
    if optional(parse_specific_char, f, b"E") or optional(
        parse_specific_char, f, b"e"
    ):
        expt = parse_int_signed(f)
    else:
        expt = b"0"
    return float(sign + bef + b"." + aft + b"E" + expt)


def read_str_f16(f):
    skip_spaces(f)
    try:
        parse_specific_string(f, "f16.nan")
        return np.float32(np.nan)
    except ValueError:
        try:
            parse_specific_string(f, "f16.inf")
            return np.float32(np.inf)
        except ValueError:
            try:
                parse_specific_string(f, "-f16.inf")
                return np.float32(-np.inf)
            except ValueError:
                x = read_str_decimal(f)
                optional_specific_string(f, "f16")
                return x


def read_str_f32(f):
    skip_spaces(f)
    try:
        parse_specific_string(f, "f32.nan")
        return np.float32(np.nan)
    except ValueError:
        try:
            parse_specific_string(f, "f32.inf")
            return np.float32(np.inf)
        except ValueError:
            try:
                parse_specific_string(f, "-f32.inf")
                return np.float32(-np.inf)
            except ValueError:
                x = read_str_decimal(f)
                optional_specific_string(f, "f32")
                return x


def read_str_f64(f):
    skip_spaces(f)
    try:
        parse_specific_string(f, "f64.nan")
        return np.float64(np.nan)
    except ValueError:
        try:
            parse_specific_string(f, "f64.inf")
            return np.float64(np.inf)
        except ValueError:
            try:
                parse_specific_string(f, "-f64.inf")
                return np.float64(-np.inf)
            except ValueError:
                x = read_str_decimal(f)
                optional_specific_string(f, "f64")
                return x


def read_str_bool(f):
    skip_spaces(f)
    if f.peek_char() == b"t":
        parse_specific_string(f, "true")
        return True
    elif f.peek_char() == b"f":
        parse_specific_string(f, "false")
        return False
    else:
        raise ValueError


def read_str_empty_array(f, type_name, rank):
    parse_specific_string(f, "empty")
    parse_specific_char(f, b"(")
    dims = []
    for i in range(rank):
        parse_specific_string(f, "[")
        dims += [int(parse_int(f))]
        parse_specific_string(f, "]")
    if np.prod(dims) != 0:
        raise ValueError
    parse_specific_string(f, type_name)
    parse_specific_char(f, b")")

    return tuple(dims)


def read_str_array_elems(f, elem_reader, type_name, rank):
    skip_spaces(f)
    try:
        parse_specific_char(f, b"[")
    except ValueError:
        return read_str_empty_array(f, type_name, rank)
    else:
        xs = sepEndBy(elem_reader, read_str_comma, f)
        skip_spaces(f)
        parse_specific_char(f, b"]")
        return xs


def read_str_array_helper(f, elem_reader, type_name, rank):
    def nested_row_reader(_):
        return read_str_array_helper(f, elem_reader, type_name, rank - 1)

    if rank == 1:
        row_reader = elem_reader
    else:
        row_reader = nested_row_reader
    return read_str_array_elems(f, row_reader, type_name, rank)


def expected_array_dims(l, rank):
    if rank > 1:
        n = len(l)
        if n == 0:
            elem = []
        else:
            elem = l[0]
        return [n] + expected_array_dims(elem, rank - 1)
    else:
        return [len(l)]


def verify_array_dims(l, dims):
    if dims[0] != len(l):
        raise ValueError
    if len(dims) > 1:
        for x in l:
            verify_array_dims(x, dims[1:])


def read_str_array(f, elem_reader, type_name, rank, bt):
    elems = read_str_array_helper(f, elem_reader, type_name, rank)
    if type(elems) == tuple:
        # Empty array
        return np.empty(elems, dtype=bt)
    else:
        dims = expected_array_dims(elems, rank)
        verify_array_dims(elems, dims)
        return np.array(elems, dtype=bt)


################################################################################

READ_BINARY_VERSION = 2

# struct format specified at
# https://docs.python.org/2/library/struct.html#format-characters


def mk_bin_scalar_reader(t):
    def bin_reader(f):
        fmt = FUTHARK_PRIMTYPES[t]["bin_format"]
        size = FUTHARK_PRIMTYPES[t]["size"]
        tf = FUTHARK_PRIMTYPES[t]["numpy_type"]
        return tf(struct.unpack("<" + fmt, f.get_chars(size))[0])

    return bin_reader


read_bin_i8 = mk_bin_scalar_reader("i8")
read_bin_i16 = mk_bin_scalar_reader("i16")
read_bin_i32 = mk_bin_scalar_reader("i32")
read_bin_i64 = mk_bin_scalar_reader("i64")

read_bin_u8 = mk_bin_scalar_reader("u8")
read_bin_u16 = mk_bin_scalar_reader("u16")
read_bin_u32 = mk_bin_scalar_reader("u32")
read_bin_u64 = mk_bin_scalar_reader("u64")

read_bin_f16 = mk_bin_scalar_reader("f16")
read_bin_f32 = mk_bin_scalar_reader("f32")
read_bin_f64 = mk_bin_scalar_reader("f64")

read_bin_bool = mk_bin_scalar_reader("bool")


def read_is_binary(f):
    skip_spaces(f)
    c = f.get_char()
    if c == b"b":
        bin_version = read_bin_u8(f)
        if bin_version != READ_BINARY_VERSION:
            panic(
                1,
                "binary-input: File uses version %i, but I only understand version %i.\n",
                bin_version,
                READ_BINARY_VERSION,
            )
        return True
    else:
        f.unget_char(c)
        return False


FUTHARK_PRIMTYPES = {
    "i8": {
        "binname": b"  i8",
        "size": 1,
        "bin_reader": read_bin_i8,
        "str_reader": read_str_i8,
        "bin_format": "b",
        "numpy_type": np.int8,
    },
    "i16": {
        "binname": b" i16",
        "size": 2,
        "bin_reader": read_bin_i16,
        "str_reader": read_str_i16,
        "bin_format": "h",
        "numpy_type": np.int16,
    },
    "i32": {
        "binname": b" i32",
        "size": 4,
        "bin_reader": read_bin_i32,
        "str_reader": read_str_i32,
        "bin_format": "i",
        "numpy_type": np.int32,
    },
    "i64": {
        "binname": b" i64",
        "size": 8,
        "bin_reader": read_bin_i64,
        "str_reader": read_str_i64,
        "bin_format": "q",
        "numpy_type": np.int64,
    },
    "u8": {
        "binname": b"  u8",
        "size": 1,
        "bin_reader": read_bin_u8,
        "str_reader": read_str_u8,
        "bin_format": "B",
        "numpy_type": np.uint8,
    },
    "u16": {
        "binname": b" u16",
        "size": 2,
        "bin_reader": read_bin_u16,
        "str_reader": read_str_u16,
        "bin_format": "H",
        "numpy_type": np.uint16,
    },
    "u32": {
        "binname": b" u32",
        "size": 4,
        "bin_reader": read_bin_u32,
        "str_reader": read_str_u32,
        "bin_format": "I",
        "numpy_type": np.uint32,
    },
    "u64": {
        "binname": b" u64",
        "size": 8,
        "bin_reader": read_bin_u64,
        "str_reader": read_str_u64,
        "bin_format": "Q",
        "numpy_type": np.uint64,
    },
    "f16": {
        "binname": b" f16",
        "size": 2,
        "bin_reader": read_bin_f16,
        "str_reader": read_str_f16,
        "bin_format": "e",
        "numpy_type": np.float16,
    },
    "f32": {
        "binname": b" f32",
        "size": 4,
        "bin_reader": read_bin_f32,
        "str_reader": read_str_f32,
        "bin_format": "f",
        "numpy_type": np.float32,
    },
    "f64": {
        "binname": b" f64",
        "size": 8,
        "bin_reader": read_bin_f64,
        "str_reader": read_str_f64,
        "bin_format": "d",
        "numpy_type": np.float64,
    },
    "bool": {
        "binname": b"bool",
        "size": 1,
        "bin_reader": read_bin_bool,
        "str_reader": read_str_bool,
        "bin_format": "b",
        "numpy_type": bool,
    },
}


def read_bin_read_type(f):
    read_binname = f.get_chars(4)

    for k, v in FUTHARK_PRIMTYPES.items():
        if v["binname"] == read_binname:
            return k
    panic(1, "binary-input: Did not recognize the type '%s'.\n", read_binname)


def numpy_type_to_type_name(t):
    for k, v in FUTHARK_PRIMTYPES.items():
        if v["numpy_type"] == t:
            return k
    raise Exception("Unknown Numpy type: {}".format(t))


def read_bin_ensure_scalar(f, expected_type):
    dims = read_bin_i8(f)

    if dims != 0:
        panic(
            1,
            "binary-input: Expected scalar (0 dimensions), but got array with %i dimensions.\n",
            dims,
        )

    bin_type = read_bin_read_type(f)
    if bin_type != expected_type:
        panic(
            1,
            "binary-input: Expected scalar of type %s but got scalar of type %s.\n",
            expected_type,
            bin_type,
        )


# ------------------------------------------------------------------------------
# General interface for reading Primitive Futhark Values
# ------------------------------------------------------------------------------


def read_scalar(f, ty):
    if read_is_binary(f):
        read_bin_ensure_scalar(f, ty)
        return FUTHARK_PRIMTYPES[ty]["bin_reader"](f)
    return FUTHARK_PRIMTYPES[ty]["str_reader"](f)


def read_array(f, expected_type, rank):
    if not read_is_binary(f):
        str_reader = FUTHARK_PRIMTYPES[expected_type]["str_reader"]
        return read_str_array(
            f,
            str_reader,
            expected_type,
            rank,
            FUTHARK_PRIMTYPES[expected_type]["numpy_type"],
        )

    bin_rank = read_bin_u8(f)

    if bin_rank != rank:
        panic(
            1,
            "binary-input: Expected %i dimensions, but got array with %i dimensions.\n",
            rank,
            bin_rank,
        )

    bin_type_enum = read_bin_read_type(f)
    if expected_type != bin_type_enum:
        panic(
            1,
            "binary-input: Expected %iD-array with element type '%s' but got %iD-array with element type '%s'.\n",
            rank,
            expected_type,
            bin_rank,
            bin_type_enum,
        )

    shape = []
    elem_count = 1
    for i in range(rank):
        bin_size = read_bin_i64(f)
        elem_count *= bin_size
        shape.append(bin_size)

    bin_fmt = FUTHARK_PRIMTYPES[bin_type_enum]["bin_format"]

    # We first read the expected number of types into a bytestring,
    # then use np.frombuffer.  This is because np.fromfile does not
    # work on things that are insufficiently file-like, like a network
    # stream.
    bytes = f.get_chars(elem_count * FUTHARK_PRIMTYPES[expected_type]["size"])
    arr = np.frombuffer(
        bytes, dtype=FUTHARK_PRIMTYPES[bin_type_enum]["numpy_type"]
    )
    arr.shape = shape

    return arr.copy()  # To ensure it is writeable.


if sys.version_info >= (3, 0):
    input_reader = ReaderInput(sys.stdin.buffer)
else:
    input_reader = ReaderInput(sys.stdin)

import re


def read_value(type_desc, reader=input_reader):
    """Read a value of the given type.  The type is a string
    representation of the Futhark type."""
    m = re.match(r"((?:\[\])*)([a-z0-9]+)$", type_desc)
    if m:
        dims = int(len(m.group(1)) / 2)
        basetype = m.group(2)
    assert m and basetype in FUTHARK_PRIMTYPES, "Unknown type: {}".format(
        type_desc
    )
    if dims > 0:
        return read_array(reader, basetype, dims)
    else:
        return read_scalar(reader, basetype)


def end_of_input(entry, f=input_reader):
    skip_spaces(f)
    if f.get_char() != b"":
        panic(1, 'Expected EOF on stdin after reading input for "%s".', entry)


def write_value_text(v, out=sys.stdout):
    if type(v) == np.uint8:
        out.write("%uu8" % v)
    elif type(v) == np.uint16:
        out.write("%uu16" % v)
    elif type(v) == np.uint32:
        out.write("%uu32" % v)
    elif type(v) == np.uint64:
        out.write("%uu64" % v)
    elif type(v) == np.int8:
        out.write("%di8" % v)
    elif type(v) == np.int16:
        out.write("%di16" % v)
    elif type(v) == np.int32:
        out.write("%di32" % v)
    elif type(v) == np.int64:
        out.write("%di64" % v)
    elif type(v) in [bool, np.bool_]:
        if v:
            out.write("true")
        else:
            out.write("false")
    elif type(v) == np.float16:
        if np.isnan(v):
            out.write("f16.nan")
        elif np.isinf(v):
            if v >= 0:
                out.write("f16.inf")
            else:
                out.write("-f16.inf")
        else:
            out.write("%.6ff16" % v)
    elif type(v) == np.float32:
        if np.isnan(v):
            out.write("f32.nan")
        elif np.isinf(v):
            if v >= 0:
                out.write("f32.inf")
            else:
                out.write("-f32.inf")
        else:
            out.write("%.6ff32" % v)
    elif type(v) == np.float64:
        if np.isnan(v):
            out.write("f64.nan")
        elif np.isinf(v):
            if v >= 0:
                out.write("f64.inf")
            else:
                out.write("-f64.inf")
        else:
            out.write("%.6ff64" % v)
    elif type(v) == np.ndarray:
        if np.prod(v.shape) == 0:
            tname = numpy_type_to_type_name(v.dtype)
            out.write(
                "empty({}{})".format(
                    "".join(["[{}]".format(d) for d in v.shape]), tname
                )
            )
        else:
            first = True
            out.write("[")
            for x in v:
                if not first:
                    out.write(", ")
                first = False
                write_value(x, out=out)
            out.write("]")
    else:
        raise Exception("Cannot print value of type {}: {}".format(type(v), v))


type_strs = {
    np.dtype("int8"): b"  i8",
    np.dtype("int16"): b" i16",
    np.dtype("int32"): b" i32",
    np.dtype("int64"): b" i64",
    np.dtype("uint8"): b"  u8",
    np.dtype("uint16"): b" u16",
    np.dtype("uint32"): b" u32",
    np.dtype("uint64"): b" u64",
    np.dtype("float16"): b" f16",
    np.dtype("float32"): b" f32",
    np.dtype("float64"): b" f64",
    np.dtype("bool"): b"bool",
}


def construct_binary_value(v):
    t = v.dtype
    shape = v.shape

    elems = 1
    for d in shape:
        elems *= d

    num_bytes = 1 + 1 + 1 + 4 + len(shape) * 8 + elems * t.itemsize
    bytes = bytearray(num_bytes)
    bytes[0] = np.int8(ord("b"))
    bytes[1] = 2
    bytes[2] = np.int8(len(shape))
    bytes[3:7] = type_strs[t]

    for i in range(len(shape)):
        bytes[7 + i * 8 : 7 + (i + 1) * 8] = np.int64(shape[i]).tobytes()

    bytes[7 + len(shape) * 8 :] = np.ascontiguousarray(v).tobytes()

    return bytes


def write_value_binary(v, out=sys.stdout):
    if sys.version_info >= (3, 0):
        out = out.buffer
    out.write(construct_binary_value(v))


def write_value(v, out=sys.stdout, binary=False):
    if binary:
        return write_value_binary(v, out=out)
    else:
        return write_value_text(v, out=out)


# End of values.py.



#!/usr/bin/env python3
#
# Turn a Futhark value encoded in the binary data format in some file
# into a PNG image in another file.  Very little error checking is
# done.  The following input types are supported:
#
#  [height][width]u8
#  [height][width]i32
#  [height][width]u32
#
#  [height][width]f32
#  [height][width]f64
#
#  [height][width][3]i8
#  [height][width][3]u8
#
# Requires purepng and Numpy.
#
# Example:
#
# $ cat input | ./some-futhark-program | tools/data2png.py /dev/stdin out.png

import sys
import struct
import numpy as np
import png

if "purepng" not in png.__file__:
    print("data2png works only with purepng, not pypng.", file=sys.stderr)
    sys.exit(1)


def read_image(f):
    # Skip binary header.
    assert f.read(1) == b"b", "Invalid binary header"
    assert f.read(1) == b"\2", "Invalid binary format"
    rank = np.int8(struct.unpack("<B", f.read(1))[0])
    type = f.read(4)

    if rank == 2 and type == b"  u8":
        height = np.int64(struct.unpack("<Q", f.read(8))[0])
        width = np.int64(struct.unpack("<Q", f.read(8))[0])
        array = np.frombuffer(f.read(height * width), dtype=np.uint8).reshape(
            height, width
        )

        return (width, height, array)

    if rank == 2 and type in [b" i32", b" u32", b" f32", b" f64"]:
        height = np.int64(struct.unpack("<Q", f.read(8))[0])
        width = np.int64(struct.unpack("<Q", f.read(8))[0])

        image = np.empty((height, width, 3), dtype=np.uint8)
        if type == b" f32":
            array = np.frombuffer(
                f.read(height * width * 4), dtype=np.float32
            ).reshape(height, width)
            image[:, :, 0] = image[:, :, 1] = image[:, :, 2] = np.uint8(
                np.float32(array) * 255
            )
        if type == b" f64":
            array = np.frombuffer(
                f.read(height * width * 8), dtype=np.float64
            ).reshape(height, width)
            image[:, :, 0] = image[:, :, 1] = image[:, :, 2] = np.uint8(
                np.float64(array) * 255
            )
        else:
            array = np.frombuffer(
                f.read(height * width * 4), dtype=np.uint32
            ).reshape(height, width)
            image[:, :, 0] = (array & 0xFF0000) >> 16
            image[:, :, 1] = (array & 0xFF00) >> 8
            image[:, :, 2] = array & 0xFF

        return (width, height, np.reshape(image, (height, width * 3)))

    elif rank == 3 and type in [b"  i8", b"  u8"]:
        height = np.int64(struct.unpack("<Q", f.read(8))[0])
        width = np.int64(struct.unpack("<Q", f.read(8))[0])
        depth = np.int64(struct.unpack("<Q", f.read(8))[0])
        assert depth == 3
        image = np.frombuffer(
            f.read(height * width * depth), dtype=np.uint8
        ).reshape(height, width, depth)

        return (width, height, np.reshape(image, (height, width * 3)))

    else:
        raise Exception(
            "Invalid rank '{}' and element type {}".format(rank, type)
        )


if __name__ == "__main__":
    infile = sys.argv[1]
    outfile = sys.argv[2]

    with open(infile, "rb") as f_in:
        width, height, img = read_image(f_in)
        with open(outfile, "wb") as f_out:
            w = png.Writer(width=width, height=height, alpha=False)
            w.write(f_out, img)



#!/usr/bin/env python3

from jsonschema import validate
import json
import sys

schema = json.load(open(sys.argv[1]))
manifest = json.load(open(sys.argv[2]))

validate(instance=manifest, schema=schema)



# Start of panic.py.


def panic(exitcode, fmt, *args):
    sys.stderr.write("%s: " % sys.argv[0])
    sys.stderr.write(fmt % args)
    sys.stderr.write("\n")
    sys.exit(exitcode)


# End of panic.py.



#!/usr/bin/env python3
#
# Construct PNG graphs of compile- and run-time performance for
# historical versions of the Futhark compiler.  The legend is almost
# useless, so manual investigation is needed to get something usable
# out of this.
#
# This program is quite naive in its construction of commands passed
# to os.system, so don't run it from a directory with interesting
# characters in its absolute path.  Run it from an initially empty
# directory, because it will create subdirectories for its own
# purposes.
#
# This program has lots of hacks and special cases to handle the
# various changes in Futhark tooling over the years.  But all things
# considered, it's not too bad.
#
# Regarding the special 'nightly' version, delete it from the
# releases/ directory if you want this script to download a newer one.
#
# $ bench-compilers.py plot_runtime 0.1.0 0.2.0 0.3.0 0.4.0 0.4.1 0.5.1 0.6.1 0.6.2 0.6.3 0.7.1 0.7.2 0.7.3 0.7.4 0.8.1 0.1 0.1 0.10.2 0.11.1 0.11.2 0.12.1 0.12.2 0.12.3 0.13.1 0.13.2 0.14.1 0.15.1 0.15.2 0.15.3 0.15.4 0.15.5 0.15.6 0.15.7 0.15.8 0.16.1 0.16.2 0.16.3 0.16.4 0.16.5 0.17.2 0.17 .3 0.18.1 0.18.2 0.18.3 0.18.4 0.18.5 0.18.6 0.19.1 0.19.2 0.19.3 0.19.4 0.19.5 nightly # nopep8
#
# The example above skips 0.9.1 because a few benchmarks do not
# validate with it.  Seems like there was a bug with certain
# reductions.


import sys
import os.path
import urllib.request
import subprocess
from pathlib import Path
import json
import time
import matplotlib.pyplot as plt
import matplotlib.ticker
import numpy as np

RELEASES_DIR = "releases"
GITS_DIR = "gits"
JSON_DIR = "json"
FUTHARK_BACKEND = "opencl"
FUTHARK_REPO = "https://github.com/diku-dk/futhark.git"
FUTHARK_BENCHMARKS_REPO = "https://github.com/diku-dk/futhark-benchmarks.git"
FUTHARK_DIR = os.path.join(GITS_DIR, "futhark")
FUTHARK_BENCHMARKS_DIR = os.path.join(GITS_DIR, "futhark-benchmarks")
BENCH_RUNS = 20

# For newer versions than this, the 'futhark-benchmarks' submodule
# points to the right revision already.
versions_to_benchmark_commits = {
    "0.3.1": "3b9c6cd06784fbf94c40a30f4302eeef119352b7",
    "0.3.0": "289000e5734705b45840ee6315bd683ab14b6ddb",
    "0.2.0": "33aabdd88bb3c935934032c5d49a7563f7078322",
    "0.1.0": "24cba07b5b70d3cbcf4195108a4caddbb649481f",
}

# Hand-curated set of benchmarks and corresponding data sets that have
# existed since version 0.1.0
benchmarks = {
    "futhark-benchmarks/accelerate/canny/canny.fut": "data/lena512.in",
    "futhark-benchmarks/accelerate/crystal/crystal.fut": '#6 ("4000i32 30.0f32 50i32 1i32 1.0f32")',
    "futhark-benchmarks/accelerate/fluid/fluid.fut": "benchmarking/medium.in",
    "futhark-benchmarks/accelerate/mandelbrot/mandelbrot.fut": '#5 ("8000i32 8000i32 -0.7f32 0.0f32 3.067f32 100i32 16....")',
    "futhark-benchmarks/accelerate/nbody/nbody.fut": "data/100000-bodies.in",
    "futhark-benchmarks/accelerate/tunnel/tunnel.fut": '#5 ("10.0f32 8000i32 8000i32")',
    "futhark-benchmarks/finpar/LocVolCalib.fut": "LocVolCalib-data/large.in",
    "futhark-benchmarks/finpar/OptionPricing.fut": "OptionPricing-data/large.in",
    "futhark-benchmarks/jgf/crypt/crypt.fut": "crypt-data/medium.in",
    "futhark-benchmarks/jgf/series/series.fut": "data/1000000.in",
    "futhark-benchmarks/misc/heston/heston32.fut": "data/100000_quotes.in",
    "futhark-benchmarks/misc/heston/heston64.fut": "data/100000_quotes.in",
    "futhark-benchmarks/misc/radix_sort/radix_sort_large.fut": "data/radix_sort_1M.in",
    "futhark-benchmarks/parboil/mri-q/mri-q.fut": "data/large.in",
    "futhark-benchmarks/parboil/sgemm/sgemm.fut": "data/medium.in",
    "futhark-benchmarks/parboil/stencil/stencil.fut": "data/default.in",
    "futhark-benchmarks/parboil/tpacf/tpacf.fut": "data/large.in",
    "futhark-benchmarks/rodinia/backprop/backprop.fut": "data/medium.in",
    "futhark-benchmarks/rodinia/bfs/bfs_filt_padded_fused.fut": "data/graph1MW_6.in",
    "futhark-benchmarks/rodinia/cfd/cfd.fut": "data/fvcorr.domn.193K.toa",
    "futhark-benchmarks/rodinia/hotspot/hotspot.fut": "data/1024.in",
    "futhark-benchmarks/rodinia/kmeans/kmeans.fut": "data/kdd_cup.in",
    "futhark-benchmarks/rodinia/lavaMD/lavaMD.fut": "data/10_boxes.in",
    "futhark-benchmarks/rodinia/lud/lud.fut": "data/2048.in",
    "futhark-benchmarks/rodinia/myocyte/myocyte.fut": "data/medium.in",
    "futhark-benchmarks/rodinia/nn/nn.fut": "data/medium.in",
    "futhark-benchmarks/rodinia/pathfinder/pathfinder.fut": "data/medium.in",
    "futhark-benchmarks/rodinia/srad/srad.fut": "data/image.in",
}


# Some data sets changed names over time - renormalise here.
def datamap(d):
    m = {
        '#6 ("4000i32 30.0f32 50i32 1i32 1.0f32")': '#5 ("4000i32 30.0f32 50i32 1i32 1.0f32")',
        '#5 ("8000i32 8000i32 -0.7f32 0.0f32 3.067f32 100i32 16....")': '#4 ("8000i32 8000i32 -0.7f32 0.0f32 3.067f32 100i32 16....")',
        '#5 ("10.0f32 8000i32 8000i32")': '#4 ("10.0f32 8000i32 8000i32")',
    }
    if d in m:
        return m[d]
    else:
        return d


def shell(cmd):
    print("$ {}".format(cmd))
    if os.system(cmd) != 0:
        raise Exception("shell command failed")


def version_before(x, y):
    if x == "nightly":
        return False
    else:
        return tuple(map(int, x.split("."))) < tuple(map(int, y.split(".")))


def runtime_json_for_version(version):
    return os.path.join(JSON_DIR, "{}.json".format(version))


def compile_json_for_version(version):
    return os.path.join(JSON_DIR, "{}-compile.json".format(version))


def release_for_version(version):
    return os.path.join(
        RELEASES_DIR, "futhark-{}-linux-x86_64".format(version)
    )


def tarball_for_version(version):
    return os.path.join(
        RELEASES_DIR, "futhark-{}-linux-x86_64.tar.xz".format(version)
    )


def futhark_compile_for_version(version):
    if version_before(version, "0.9.1"):
        return os.path.abspath(
            os.path.join(
                release_for_version(version),
                "bin",
                "futhark-{}".format(FUTHARK_BACKEND),
            )
        )
    else:
        return os.path.join(
            release_for_version(version),
            "bin/futhark {}".format(FUTHARK_BACKEND),
        )


def futhark_bench_for_version(version):
    bs = " ".join(list(benchmarks.keys()))
    if version_before(version, "0.9.1"):
        prog = os.path.abspath(
            os.path.join(release_for_version(version), "bin", "futhark-bench")
        )
        compiler = os.path.abspath(
            os.path.join(
                release_for_version(version),
                "bin",
                "futhark-{}".format(FUTHARK_BACKEND),
            )
        )
        return "{} -r {} --compiler={} {}".format(
            prog, BENCH_RUNS, compiler, bs
        )
    else:
        prog = os.path.abspath(
            os.path.join(release_for_version(version), "bin", "futhark")
        )
        return "{} bench -r {} --backend={} --futhark={} {}".format(
            prog, BENCH_RUNS, FUTHARK_BACKEND, prog, bs
        )


def tarball_url(version):
    return "https://futhark-lang.org/releases/futhark-{}-linux-x86_64.tar.xz".format(
        version
    )


def ensure_tarball(version):
    f = tarball_for_version(version)
    if not os.path.exists(f):
        print("Downloading {}...".format(tarball_url(version)))
        urllib.request.urlretrieve(tarball_url(version), f)


def ensure_release(version):
    if not os.path.exists(release_for_version(version)):
        ensure_tarball(version)
        print("Extracting {}...".format(tarball_for_version(version)))
        shell(
            "cd {} && tar xf ../{}".format(
                RELEASES_DIR, tarball_for_version(version)
            )
        )


def ensure_repo(what, url):
    dir = os.path.join(GITS_DIR, what)
    if os.path.exists(dir):
        shell("cd {} && git checkout master && git pull".format(dir))
    else:
        shell("git clone --recursive {} {}".format(url, dir))
    return dir


def ensure_futhark_repo():
    return ensure_repo("futhark", FUTHARK_REPO)


def ensure_benchmarks(version):
    if version in versions_to_benchmark_commits:
        ensure_repo("futhark-benchmarks", FUTHARK_BENCHMARKS_REPO)
        shell(
            "cd {} && git checkout -f {}".format(
                FUTHARK_BENCHMARKS_DIR, versions_to_benchmark_commits[version]
            )
        )
        return GITS_DIR
    else:
        d = ensure_futhark_repo()
        tag = "master" if version == "nightly" else "v{}".format(version)
        shell(
            "cd {} && git checkout -f {} && git submodule update".format(
                d, tag
            )
        )
        return os.path.join(d)


def produce_runtime_json(version):
    f = runtime_json_for_version(version)
    ensure_release(version)
    ensure_repo("futhark", FUTHARK_REPO)
    d = ensure_benchmarks(version)
    json_f = os.path.abspath(runtime_json_for_version(version))
    shell(
        "cd {} && {} --json={}".format(
            d, futhark_bench_for_version(version), json_f
        )
    )


def ensure_runtime_json(version):
    f = runtime_json_for_version(version)
    if not os.path.exists(f):
        produce_runtime_json(version)
    return f


def produce_compile_json(version):
    ensure_release(version)
    d = ensure_benchmarks(version)
    compiletimes = {}
    for b in benchmarks:
        bef = time.time()
        shell(
            "{} {}".format(
                futhark_compile_for_version(version), os.path.join(d, b)
            )
        )
        aft = time.time()
        compiletimes[b] = aft - bef
    with open(compile_json_for_version(version), "w") as f:
        json.dump(compiletimes, f)


def ensure_compile_json(version):
    f = compile_json_for_version(version)
    if not os.path.exists(f):
        produce_compile_json(version)
    return f


def ensure_json(version):
    ensure_runtime_json(version)
    ensure_compile_json(version)


def load_json(f):
    return json.load(open(f, "r"))


def plot_runtime(versions):
    fastest = {}
    results_per_bench = {}
    for v in versions:
        results = load_json(ensure_runtime_json(v))
        for b, d in benchmarks.items():
            try:
                x = np.median(results[b]["datasets"][d]["runtimes"])
            except KeyError:
                x = np.median(results[b]["datasets"][datamap(d)]["runtimes"])

            if (b, d) in fastest:
                fastest[b, d] = min(fastest[b, d], x)
            else:
                fastest[b, d] = x

            if (b, d) in results_per_bench:
                results_per_bench[b, d] += [x]
            else:
                results_per_bench[b, d] = [x]

    for b, d in benchmarks.items():
        results_per_bench[b, d] = (
            np.array(results_per_bench[b, d]) / fastest[b, d]
        )

    fig, ax = plt.subplots()

    ax.set_yscale("log")
    ax.yaxis.set_major_formatter(matplotlib.ticker.ScalarFormatter())
    ax.set_ylabel("Runtime relative to fastest")
    plt.xticks(rotation=90)

    for b, d in benchmarks.items():
        xs = versions + ["..."]
        ys = np.append(results_per_bench[b, d], [results_per_bench[b, d][-1]])
        ax.plot(xs, ys, label=b[len("futhark-benchmarks/") :])

    ax.grid()

    ax.legend(
        bbox_to_anchor=(0, -0.35), ncol=3, loc="center left", fontsize=5.5
    )
    fig.savefig("runtime.png", dpi=300, figsize=(12, 6), bbox_inches="tight")


def plot_compiletime(versions):
    fastest = {}
    results_per_bench = {}
    for v in versions:
        results = load_json(ensure_compile_json(v))
        for b in benchmarks:
            x = np.mean(results[b])

            if b in fastest:
                fastest[b] = min(fastest[b], x)
            else:
                fastest[b] = x

            if b in results_per_bench:
                results_per_bench[b] += [x]
            else:
                results_per_bench[b] = [x]

    for b in benchmarks:
        results_per_bench[b] = np.array(results_per_bench[b]) / fastest[b]

    fig, ax = plt.subplots()

    ax.set_yscale("log")
    ax.yaxis.set_major_formatter(matplotlib.ticker.ScalarFormatter())
    ax.set_ylabel("Compilation time relative to fastest")
    plt.xticks(rotation=90)

    for b in benchmarks:
        xs = versions + ["..."]
        ys = np.append(results_per_bench[b], [results_per_bench[b][-1]])
        ax.plot(xs, ys, label=b[len("futhark-benchmarks/") :])

    ax.grid()

    ax.legend(
        bbox_to_anchor=(0, -0.35), ncol=3, loc="center left", fontsize=5.5
    )
    fig.savefig(
        "compiletime.png", dpi=300, figsize=(12, 6), bbox_inches="tight"
    )


if __name__ == "__main__":
    _, cmd, *args = sys.argv

    Path(RELEASES_DIR).mkdir(parents=True, exist_ok=True)
    Path(GITS_DIR).mkdir(parents=True, exist_ok=True)
    Path(JSON_DIR).mkdir(parents=True, exist_ok=True)

    if cmd == "bench":
        ensure_json(*args)
    elif cmd == "benchdir":
        print(os.path.join(ensure_benchmarks(*args), "futhark-benchmarks"))
    elif cmd == "plot_runtime":
        plot_runtime(args)
    elif cmd == "plot_compiletime":
        plot_compiletime(args)
    else:
        print("Invalid command: {}\n".format(cmd))
        sys.exit(1)



#!/usr/bin/env python3
#
# Usage: let2def.py prog.fut
#
# Replaces top-level 'let's with 'def'.

from subprocess import check_output
import sys
import re
import os

prog = sys.argv[1]
out = check_output(["futhark", "defs", prog]).decode("utf-8")

funlocs = set()
for row in out.split("\n"):
    m = re.match("value [^ ]+ ([^:]+):([0-9]+):([0-9]+)", row)
    if m and m[1] == prog:
        funlocs.add((int(m[2]), int(m[3])))

f = open(prog, "r+")
s = list(f.read())

line = 1
col = 0
n = len(s)
for i in range(n):
    if s[i] == "\n":
        line += 1
        col = 0
    elif s[i] == "\t":
        # Futhark lexer assumes tabwidth=8 for source positions.
        col += 1
        while col % 8 != 0:
            col += 1
    else:
        col += 1
    if s[i : i + 3] == list("let") and (line, col) in funlocs:
        s[i : i + 3] = "def"
f.seek(0)
f.truncate()
f.write("".join(s))



# Start of server.py

import sys
import time
import shlex  # For string splitting


class Server:
    def __init__(self, ctx):
        self._ctx = ctx
        self._vars = {}

    class Failure(BaseException):
        def __init__(self, msg):
            self.msg = msg

    def _get_arg(self, args, i):
        if i < len(args):
            return args[i]
        else:
            raise self.Failure("Insufficient command args")

    def _get_entry_point(self, entry):
        if entry in self._ctx.entry_points:
            return self._ctx.entry_points[entry]
        else:
            raise self.Failure("Unknown entry point: %s" % entry)

    def _check_var(self, vname):
        if not vname in self._vars:
            raise self.Failure("Unknown variable: %s" % vname)

    def _check_new_var(self, vname):
        if vname in self._vars:
            raise self.Failure("Variable already exists: %s" % vname)

    def _get_var(self, vname):
        self._check_var(vname)
        return self._vars[vname]

    def _cmd_inputs(self, args):
        entry = self._get_arg(args, 0)
        for t in self._get_entry_point(entry)[1]:
            print(t)

    def _cmd_outputs(self, args):
        entry = self._get_arg(args, 0)
        for t in self._get_entry_point(entry)[2]:
            print(t)

    def _cmd_dummy(self, args):
        pass

    def _cmd_free(self, args):
        for vname in args:
            self._check_var(vname)
            del self._vars[vname]

    def _cmd_rename(self, args):
        oldname = self._get_arg(args, 0)
        newname = self._get_arg(args, 1)
        self._check_var(oldname)
        self._check_new_var(newname)
        self._vars[newname] = self._vars[oldname]
        del self._vars[oldname]

    def _cmd_call(self, args):
        entry = self._get_entry_point(self._get_arg(args, 0))
        entry_fname = entry[0]
        num_ins = len(entry[1])
        num_outs = len(entry[2])
        exp_len = 1 + num_outs + num_ins

        if len(args) != exp_len:
            raise self.Failure("Invalid argument count, expected %d" % exp_len)

        out_vnames = args[1 : num_outs + 1]

        for out_vname in out_vnames:
            self._check_new_var(out_vname)

        in_vnames = args[1 + num_outs :]
        ins = [self._get_var(in_vname) for in_vname in in_vnames]

        try:
            (runtime, vals) = getattr(self._ctx, entry_fname)(*ins)
        except Exception as e:
            raise self.Failure(str(e))

        print("runtime: %d" % runtime)

        if num_outs == 1:
            self._vars[out_vnames[0]] = vals
        else:
            for out_vname, val in zip(out_vnames, vals):
                self._vars[out_vname] = val

    def _store_val(self, f, value):
        # In case we are using the PyOpenCL backend, we first
        # need to convert OpenCL arrays to ordinary NumPy
        # arrays.  We do this in a nasty way.
        if isinstance(value, opaque):
            for component in value.data:
                self._store_val(f, component)
        elif (
            isinstance(value, np.number)
            or isinstance(value, bool)
            or isinstance(value, np.bool_)
            or isinstance(value, np.ndarray)
        ):
            # Ordinary NumPy value.
            f.write(construct_binary_value(value))
        else:
            # Assuming PyOpenCL array.
            f.write(construct_binary_value(value.get()))

    def _cmd_store(self, args):
        fname = self._get_arg(args, 0)

        with open(fname, "wb") as f:
            for i in range(1, len(args)):
                self._store_val(f, self._get_var(args[i]))

    def _restore_val(self, reader, typename):
        if typename in self._ctx.opaques:
            vs = []
            for t in self._ctx.opaques[typename]:
                vs += [read_value(t, reader)]
            return opaque(typename, *vs)
        else:
            return read_value(typename, reader)

    def _cmd_restore(self, args):
        if len(args) % 2 == 0:
            raise self.Failure("Invalid argument count")

        fname = args[0]
        args = args[1:]

        with open(fname, "rb") as f:
            reader = ReaderInput(f)
            while args != []:
                vname = args[0]
                typename = args[1]
                args = args[2:]

                if vname in self._vars:
                    raise self.Failure("Variable already exists: %s" % vname)

                try:
                    self._vars[vname] = self._restore_val(reader, typename)
                except ValueError:
                    raise self.Failure(
                        "Failed to restore variable %s.\n"
                        "Possibly malformed data in %s.\n" % (vname, fname)
                    )

            skip_spaces(reader)
            if reader.get_char() != b"":
                raise self.Failure("Expected EOF after reading values")

    def _cmd_types(self, args):
        for k in self._ctx.opaques.keys():
            print(k)

    def _cmd_entry_points(self, args):
        for k in self._ctx.entry_points.keys():
            print(k)

    _commands = {
        "inputs": _cmd_inputs,
        "outputs": _cmd_outputs,
        "call": _cmd_call,
        "restore": _cmd_restore,
        "store": _cmd_store,
        "free": _cmd_free,
        "rename": _cmd_rename,
        "clear": _cmd_dummy,
        "pause_profiling": _cmd_dummy,
        "unpause_profiling": _cmd_dummy,
        "report": _cmd_dummy,
        "types": _cmd_types,
        "entry_points": _cmd_entry_points,
    }

    def _process_line(self, line):
        lex = shlex.shlex(line)
        lex.quotes = '"'
        lex.whitespace_split = True
        lex.commenters = ""
        words = list(lex)
        if words == []:
            raise self.Failure("Empty line")
        else:
            cmd = words[0]
            args = words[1:]
            if cmd in self._commands:
                self._commands[cmd](self, args)
            else:
                raise self.Failure("Unknown command: %s" % cmd)

    def run(self):
        while True:
            print("%%% OK", flush=True)
            line = sys.stdin.readline()
            if line == "":
                return
            try:
                self._process_line(line)
            except self.Failure as e:
                print("%%% FAILURE")
                print(e.msg)


# End of server.py
